import datetime as dt
from unittest.mock import patch

from bravado.exception import HTTPBadGateway, HTTPInternalServerError
from pytz import utc

from django.test import override_settings
from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EvePlanet

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import BravadoResponseStub, NoSocketsTestCase, queryset_pks

from structures.core.notification_types import NotificationType
from structures.models import Notification, Structure, StructureItem
from structures.tests.testdata.factories import (
    EveCharacterFactory,
    EveCorporationInfoFactory,
    EveEntityCorporationFactory,
    JumpFuelAlertConfigFactory,
    OwnerFactory,
    SkyhookFactory,
    StarbaseFactory,
    StructureFactory,
    StructureItemFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
    datetime_to_esi,
)
from structures.tests.testdata.helpers import (
    NearestCelestial,
    load_eve_entities,
    load_notification_entities,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

OWNERS_PATH = "structures.models.owners"
NOTIFICATIONS_PATH = "structures.models.notifications"


@patch(OWNERS_PATH + ".esi")
class TestFetchNotificationsEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.sender_corporation = EveEntityCorporationFactory()
        cls.character = EveCharacterFactory()
        cls.user = UserMainDefaultOwnerFactory(main_character__character=cls.character)
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_notifications",
                "character_id",
                needs_token=True,
                data={
                    f"{cls.character.character_id}": [
                        {
                            "notification_id": 1000000505,
                            "type": "StructureLostShields",
                            "sender_id": cls.sender_corporation.id,
                            "sender_type": "corporation",
                            "timestamp": "2019-10-04 14:52:00",
                            "text": "solarsystemID: 30002537\nstructureID: &id001 1000000000001\nstructureShowInfoData:\n- showinfo\n- 35832\n- *id001\nstructureTypeID: 35832\ntimeLeft: 1727805401093\ntimestamp: 132148470780000000\nvulnerableTime: 9000000000\n",
                        },
                    ]
                },
            )
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(endpoints)

    @patch(OWNERS_PATH + ".notify", spec=True)
    @patch(OWNERS_PATH + ".now", spec=True)
    def test_should_inform_user_about_successful_update(
        self, mock_now, mock_notify, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_now.return_value = dt.datetime(2019, 8, 16, 14, 15, tzinfo=utc)
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        StructureFactory(owner=owner, id=1000000000001)
        # when
        owner.fetch_notifications_esi(self.user)
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        self.assertTrue(mock_notify.called)

    def test_should_create_notifications(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        StructureFactory(owner=owner, id=1000000000001)
        # when
        owner.fetch_notifications_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        # should only contain the right notifications
        notif_ids_current = set(
            Notification.objects.values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notif_ids_current, {1000000505})

    @patch(OWNERS_PATH + ".now")
    def test_should_set_moon_for_structure_if_missing(self, mock_now, mock_esi_client):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_notifications",
                "character_id",
                needs_token=True,
                data={
                    f"{character.character_id}": [
                        {
                            "notification_id": 1000000404,
                            "type": "MoonminingExtractionStarted",
                            "sender_id": self.sender_corporation.id,
                            "sender_type": "corporation",
                            "timestamp": "2019-11-13 23:33:00",
                            "text": 'autoTime: 132186924601059151\nmoonID: 40161465\noreVolumeByType:\n  46300: 1288475.124715103\n  46301: 544691.7637724016\n  46302: 526825.4047522942\n  46303: 528996.6386983792\nreadyTime: 132186816601059151\nsolarSystemID: 30002537\nstartedBy: 1001\nstartedByLink: <a href="showinfo:1383//1001">Bruce Wayne</a>\nstructureID: 1000000000002\nstructureLink: <a href="showinfo:35835//1000000000002">Dummy</a>\nstructureName: Dummy\nstructureTypeID: 35835\n',
                            "is_read": False,
                        },
                    ]
                },
            )
        ]
        mock_esi_client.client = EsiClientStub.create_from_endpoints(endpoints)
        mock_now.return_value = dt.datetime(2019, 11, 13, 23, 50, 0, tzinfo=utc)
        owner = OwnerFactory(
            user=user, notifications_last_update_at=None, characters=[character]
        )
        structure = StructureFactory(owner=owner, id=1000000000002, eve_type_id=35835)
        # when
        owner.fetch_notifications_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        structure.refresh_from_db()
        self.assertEqual(structure.eve_moon_id, 40161465)

    def test_report_error_when_esi_returns_error_during_sync(self, mock_esi):
        def my_callback(*args, **kwargs):
            raise HTTPBadGateway(
                BravadoResponseStub(status_code=502, reason="Test Exception")
            )

        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_notifications",
                "character_id",
                needs_token=True,
                data=[],
                side_effect=my_callback,
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        StructureFactory(owner=owner, id=1000000000001)
        # when
        with self.assertRaises(HTTPBadGateway):
            owner.fetch_notifications_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_notification_sync_fresh)

    def test_should_create_notifications_from_scratch(self, mock_esi):
        # given
        owner = OwnerFactory(notifications_last_update_at=None)
        sender = EveEntityCorporationFactory()
        eve_character = owner.characters.first().character_ownership.character
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_notifications",
                "character_id",
                needs_token=True,
                data={
                    str(eve_character.character_id): [
                        {
                            "notification_id": 42,
                            "is_read": False,
                            "sender_id": sender.id,
                            "sender_type": "corporation",
                            "text": "{}\n",
                            "timestamp": datetime_to_esi(now()),
                            "type": "CorpBecameWarEligible",
                        }
                    ]
                },
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        owner.fetch_notifications_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        self.assertEqual(owner.notification_set.count(), 1)
        obj = owner.notification_set.first()
        self.assertEqual(
            obj.notif_type, NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
        )

    def test_should_handle_other_sender_correctly(self, mock_esi):
        # given
        owner = OwnerFactory(notifications_last_update_at=None)
        eve_character = owner.characters.first().character_ownership.character
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_notifications",
                "character_id",
                needs_token=True,
                data={
                    str(eve_character.character_id): [
                        {
                            "notification_id": 42,
                            "is_read": False,
                            "sender_id": 1,
                            "sender_type": "other",
                            "text": "{}\n",
                            "timestamp": datetime_to_esi(now()),
                            "type": "CorpBecameWarEligible",
                        }
                    ]
                },
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        owner.fetch_notifications_esi()
        # then
        obj = owner.notification_set.get(notification_id=42)
        self.assertIsNone(obj.sender)


@override_settings(DEBUG=True)
@patch(NOTIFICATIONS_PATH + ".STRUCTURES_REPORT_NPC_ATTACKS", True)
@patch(NOTIFICATIONS_PATH + ".Webhook.send_message", spec=True)
class TestSendNewNotifications(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory(
            is_alliance_main=True, webhooks=False, forwarding_last_update_at=None
        )
        load_notification_entities(cls.owner)

    def setUp(self) -> None:
        self.owner.webhooks.clear()

    # TODO: Temporarily disabled
    # @patch(
    #     NOTIFICATIONS_PATH + ".STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS", False
    # )
    def test_should_send_all_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(notification_types=NotificationType.values)
        self.owner.webhooks.add(webhook)

        # when
        self.owner.send_new_notifications()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            self.owner.notification_set.filter(
                notif_type__in=NotificationType.values
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)

    # TODO: temporary disabled
    # @patch(
    #     NOTIFICATIONS_PATH + ".STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS", True
    # )
    # def test_should_send_all_notifications_except_fuel_alerts(self, mock_send_message):
    #     # given
    #     mock_send_message.return_value = True
    #     self.user = AuthUtils.add_permission_to_user_by_name(
    #         "structures.add_structure_owner", self.user
    #     )
    #     # when
    #     self.self.owner.send_new_notifications()
    #     # then
    #     self.self.owner.refresh_from_db()
    #     self.assertTrue(self.self.owner.is_forwarding_sync_fresh)
    #     notifications_processed = {
    #         int(args[1]["embeds"][0].footer.text[-10:])
    #         for args in mock_send_message.call_args_list
    #     }
    #     notif_types = set(NotificationType.values)
    #     notif_types.discard(NotificationType.STRUCTURE_FUEL_ALERT)
    #     notif_types.discard(NotificationType.TOWER_RESOURCE_ALERT_MSG)
    #     notifications_expected = set(
    #         self.self.owner.notifications.filter(notif_type__in=notif_types).values_list(
    #             "notification_id", flat=True
    #         )
    #     )
    #     self.assertSetEqual(notifications_processed, notifications_expected)

    def test_should_send_all_notifications_corp(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        corporation = EveCorporationInfoFactory(corporation_id=2011)
        character = EveCharacterFactory(corporation=corporation)
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(
            user=user,
            is_alliance_main=True,
            webhooks__notification_types=NotificationType.values,
            forwarding_last_update_at=None,
        )
        load_notification_entities(owner)

        # when
        owner.send_new_notifications()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            owner.notification_set.filter(
                notif_type__in=NotificationType.values
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)

    def test_should_only_send_selected_notification_types(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        selected_notif_types = [
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.STRUCTURE_DESTROYED,
        ]
        webhook = WebhookFactory(notification_types=selected_notif_types)
        self.owner.webhooks.add(webhook)

        # when
        self.owner.send_new_notifications()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            Notification.objects.filter(
                notif_type__in=selected_notif_types
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)


@patch(OWNERS_PATH + ".esi")
class TestOwnerUpdateAssetEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.corporation = EveCharacterFactory()
        character = EveCharacterFactory(corporation=cls.corporation)
        cls.user = UserMainDefaultOwnerFactory(main_character__character=character)

        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    f"{cls.corporation.corporation_id}": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000001001,
                            "location_flag": "QuantumCoreRoom",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 56201,
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000001002,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000002001,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000002,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1500000000001,
                            "location_flag": "AutoFit",
                            "location_id": 30002537,  # Amamake,
                            "location_type": "solar_system",
                            "quantity": 1,
                            "type_id": 16213,  # control tower
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1500000000002,
                            "location_flag": "AutoFit",
                            "location_id": 30002537,  # Amamake,
                            "location_type": "solar_system",
                            "quantity": 1,
                            "type_id": 32226,
                        },
                    ],
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    f"{cls.corporation.corporation_id}": [
                        {"item_id": 1500000000002, "position": {"x": 1, "y": 2, "z": 3}}
                    ]
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(endpoints)

    def test_should_update_upwell_items_for_owner(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001)
        StructureFactory(owner=owner, id=1000000000002)
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_assets_sync_fresh)
        self.assertSetEqual(
            queryset_pks(StructureItem.objects.all()),
            {1300000001001, 1300000001002, 1300000002001},
        )
        obj = owner.structures.get(pk=1000000000001).items.get(pk=1300000001001)
        self.assertEqual(obj.eve_type_id, 56201)
        self.assertEqual(
            obj.location_flag, StructureItem.LocationFlag.QUANTUM_CORE_ROOM
        )
        self.assertEqual(obj.quantity, 1)
        self.assertFalse(obj.is_singleton)

        obj = owner.structures.get(pk=1000000000001).items.get(pk=1300000001002)
        self.assertEqual(obj.eve_type_id, 35894)
        self.assertEqual(obj.location_flag, "ServiceSlot0")
        self.assertEqual(obj.quantity, 1)
        self.assertTrue(obj.is_singleton)

        structure = owner.structures.get(id=1000000000001)
        self.assertTrue(structure.has_fitting)
        self.assertTrue(structure.has_core)

        structure = owner.structures.get(id=1000000000002)
        self.assertTrue(structure.has_fitting)
        self.assertFalse(structure.has_core)

    @patch(OWNERS_PATH + ".notify", spec=True)
    def test_should_inform_user_about_successful_update(self, mock_notify, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001)
        # when
        owner.update_asset_esi(user=self.user)
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_assets_sync_fresh)
        self.assertTrue(mock_notify.called)

    def test_should_raise_exception_if_esi_has_error(self, mock_esi):
        def my_callback(**kwargs):
            raise HTTPInternalServerError(
                BravadoResponseStub(status_code=500, reason="Test")
            )

        # given
        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                side_effect=my_callback,
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001)
        # when
        with self.assertRaises(HTTPInternalServerError):
            owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_assets_sync_fresh)

    def test_should_remove_assets_that_no_longer_exist_for_existing_structure(
        self, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        structure = StructureFactory(owner=owner, id=1000000000001)
        item = StructureItemFactory(structure=structure)
        # when
        owner.update_asset_esi()
        # then
        self.assertFalse(structure.items.filter(pk=item.pk).exists())

    def test_should_remove_assets_that_no_longer_exist_for_removed_structure(
        self, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001)
        structure = StructureFactory(owner=owner, id=1000000000666)
        item = StructureItemFactory(structure=structure)
        # when
        owner.update_asset_esi()
        # then
        self.assertFalse(structure.items.filter(pk=item.pk).exists())

    def test_should_handle_asset_moved_to_another_structure(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        structure_1 = StructureFactory(owner=owner, id=1000000000001)
        structure_2 = StructureFactory(owner=owner, id=1000000000002)
        StructureItemFactory(
            structure=structure_2,
            id=1300000001002,
            eve_type_id=35894,
            location_flag="ServiceSlot0",
            is_singleton=True,
            quantity=1,
        )
        # when
        owner.update_asset_esi()
        # then
        self.assertSetEqual(queryset_pks(structure_2.items.all()), {1300000002001})
        self.assertSetEqual(
            queryset_pks(structure_1.items.all()), {1300000001001, 1300000001002}
        )

    def test_should_not_delete_assets_from_other_owners(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        user_2 = UserMainDefaultOwnerFactory()
        owner_2 = OwnerFactory(user=user_2)
        structure_2 = StructureFactory(
            owner=owner_2, id=1000000000004, quantum_core=False
        )
        StructureItemFactory(structure=structure_2, id=1300000003001)

        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000001, quantum_core=False)
        StructureFactory(owner=owner, id=1000000000002, quantum_core=False)

        # when
        owner.update_asset_esi()

        # then
        self.assertSetEqual(
            queryset_pks(StructureItem.objects.all()),
            {1300000001001, 1300000001002, 1300000002001, 1300000003001},
        )

    def test_should_remove_outdated_jump_fuel_alerts(self, mock_esi):
        # given
        user = UserMainDefaultOwnerFactory()
        owner = OwnerFactory(user=user)
        structure = StructureFactory(owner=owner, id=1000000000004)
        config = JumpFuelAlertConfigFactory(threshold=100)
        structure.jump_fuel_alerts.create(structure=structure, config=config)

        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    f"{owner.corporation.corporation_id}": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000003001,
                            "location_flag": "StructureFuel",
                            "location_id": 1000000000004,
                            "location_type": "item",
                            "quantity": 5000,
                            "type_id": 16273,
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={f"{owner.corporation.corporation_id}": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        owner.update_asset_esi()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    # TODO: Add tests for error cases

    def test_should_update_starbase_items_for_owner(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        StarbaseFactory(
            owner=owner, id=1500000000001, position_x=1, position_y=2, position_z=3
        )  # position needed to match assets
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_assets_sync_fresh)
        self.assertSetEqual(queryset_pks(StructureItem.objects.all()), {1500000000002})

    def test_should_update_upwell_items_for_owner_with_invalid_locations(
        self, mock_esi
    ):
        # given
        user = UserMainDefaultOwnerFactory()
        owner = OwnerFactory(user=user)
        structure = StructureFactory(owner=owner, id=1000000000004)

        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    f"{owner.corporation.corporation_id}": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000003001,
                            "location_flag": "StructureFuel",
                            "location_id": 1000000000004,
                            "location_type": "item",
                            "quantity": 5000,
                            "type_id": 16273,
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                http_error_code=404,
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        owner.update_asset_esi()
        # then
        self.assertTrue(structure.items.filter(id=1300000003001).exists())


@patch(OWNERS_PATH + ".STRUCTURES_FEATURE_SKYHOOKS", True)
@patch(OWNERS_PATH + ".EveSolarSystem.nearest_celestial")
@patch(OWNERS_PATH + ".esi")
class TestOwnerUpdateSkyhooks(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.corporation = EveCharacterFactory()
        character = EveCharacterFactory(corporation=cls.corporation)
        cls.user = UserMainDefaultOwnerFactory(main_character__character=character)
        cls.planet = EvePlanet.objects.get(id=40161469)
        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    f"{cls.corporation.corporation_id}": [
                        {
                            "is_singleton": True,
                            "item_id": 1000000010001,
                            "location_flag": "AutoFit",
                            "location_id": 30002537,
                            "location_type": "solar_system",
                            "quantity": 1,
                            "type_id": 81080,
                        },
                    ],
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    f"{cls.corporation.corporation_id}": [
                        {"item_id": 1000000010001, "position": {"x": 1, "y": 2, "z": 3}}
                    ]
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(endpoints)

    def test_should_create_new_skyhooks_from_scratch(
        self, mock_esi, mock_nearest_celestial
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_nearest_celestial.return_value = NearestCelestial(
            eve_object=self.planet, distance=35_000_000, eve_type=self.planet.eve_type
        )
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertEqual(owner.structures.count(), 1)
        obj: Structure = owner.structures.get(pk=1000000010001)
        self.assertTrue(obj.is_skyhook)
        self.assertEqual(obj.eve_planet, self.planet)

    def test_should_remove_obsolete_skyhooks(self, mock_esi, mock_nearest_celestial):
        # given
        mock_esi.client = self.esi_client_stub
        mock_nearest_celestial.return_value = NearestCelestial(
            eve_object=self.planet, distance=35_000_000, eve_type=self.planet.eve_type
        )
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        SkyhookFactory.create(owner=owner)
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertEqual(owner.structures.count(), 1)
        obj: Structure = owner.structures.get(pk=1000000010001)
        self.assertTrue(obj.is_skyhook)
        self.assertEqual(obj.eve_planet, self.planet)

    def test_should_update_existing_skyhook(self, mock_esi, mock_nearest_celestial):
        # given
        mock_esi.client = self.esi_client_stub
        mock_nearest_celestial.return_value = NearestCelestial(
            eve_object=self.planet, distance=35_000_000, eve_type=self.planet.eve_type
        )
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        SkyhookFactory.create(owner=owner, id=1000000010001, eve_planet_name="Thera I")
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertEqual(owner.structures.count(), 1)
        obj: Structure = owner.structures.get(pk=1000000010001)
        self.assertTrue(obj.is_skyhook)
        self.assertEqual(obj.eve_planet, self.planet)

    def test_should_ignore_os_error_when_resolving_planet(
        self, mock_esi, mock_nearest_celestial
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_nearest_celestial.side_effect = OSError
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertEqual(owner.structures.count(), 1)
        obj: Structure = owner.structures.get(pk=1000000010001)
        self.assertTrue(obj.is_skyhook)
        self.assertIsNone(obj.eve_planet)

    def test_should_ignore_no_reply_when_resolving_planet(
        self, mock_esi, mock_nearest_celestial
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_nearest_celestial.return_value = None
        owner = OwnerFactory(user=self.user, assets_last_update_at=None)
        # when
        owner.update_asset_esi()
        # then
        owner.refresh_from_db()
        self.assertEqual(owner.structures.count(), 1)
        obj: Structure = owner.structures.get(pk=1000000010001)
        self.assertTrue(obj.is_skyhook)
        self.assertIsNone(obj.eve_planet)


class TestOwnerToken(NoSocketsTestCase):
    def test_should_return_valid_token(self):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        # when
        token = owner.characters.first().valid_token()
        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertEqual(token.character_id, character.character_id)

    def test_should_return_none_if_no_valid_token_found(self):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        user.token_set.first().scopes.clear()
        # when
        token = owner.characters.first().valid_token()
        # then
        self.assertIsNone(token)


@patch(OWNERS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", True)
@patch(OWNERS_PATH + ".notify_admins")
class TestOwnerUpdateIsUp(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.corporation = EveCorporationInfoFactory()

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_do_nothing_when_still_up(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=True, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertFalse(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", False)
    def test_should_report_when_down(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=True, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertFalse(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertFalse(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", False)
    def test_should_not_report_again_when_still_down(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=False, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertFalse(result)
        self.assertFalse(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertFalse(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_report_when_up_again(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=False, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_report_when_up_for_the_first_time(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=None, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)



from unittest.mock import patch

from django.test import TestCase

from structures.core.notification_types import NotificationType

MODULE_PATH = "structures.core.notification_types"


class TestNotificationType(TestCase):
    def test_should_return_enabled_values_only(self):
        # when
        with patch(MODULE_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", False):
            values = NotificationType.values_enabled()
        # then
        self.assertNotIn(NotificationType.STRUCTURE_REFUELED_EXTRA, values)
        self.assertNotIn(NotificationType.TOWER_REFUELED_EXTRA, values)

    def test_should_return_all_values(self):
        # when
        with patch(MODULE_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True):
            values = NotificationType.values_enabled()
        # then
        self.assertIn(NotificationType.STRUCTURE_REFUELED_EXTRA, values)
        self.assertIn(NotificationType.TOWER_REFUELED_EXTRA, values)

    def test_should_return_enabled_choices_only(self):
        # when
        with patch(MODULE_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", False):
            choices = NotificationType.choices_enabled()
        # then
        types = {choice[0] for choice in choices}
        self.assertNotIn(NotificationType.STRUCTURE_REFUELED_EXTRA, types)
        self.assertNotIn(NotificationType.TOWER_REFUELED_EXTRA, types)

    def test_should_return_all_choices(self):
        # when
        with patch(MODULE_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True):
            choices = NotificationType.choices_enabled()
        # then
        types = {choice[0] for choice in choices}
        self.assertIn(NotificationType.STRUCTURE_REFUELED_EXTRA, types)
        self.assertIn(NotificationType.TOWER_REFUELED_EXTRA, types)

    def test_has_correct_esi_values(self):
        # given
        esi_valid_notification_types = {
            "AcceptedAlly",
            "AcceptedSurrender",
            "AgentRetiredTrigravian",
            "AllAnchoringMsg",
            "AllMaintenanceBillMsg",
            "AllStrucInvulnerableMsg",
            "AllStructVulnerableMsg",
            "AllWarCorpJoinedAllianceMsg",
            "AllWarDeclaredMsg",
            "AllWarInvalidatedMsg",
            "AllWarRetractedMsg",
            "AllWarSurrenderMsg",
            "AllianceCapitalChanged",
            "AllianceWarDeclaredV2",
            "AllyContractCancelled",
            "AllyJoinedWarAggressorMsg",
            "AllyJoinedWarAllyMsg",
            "AllyJoinedWarDefenderMsg",
            "BattlePunishFriendlyFire",
            "BillOutOfMoneyMsg",
            "BillPaidCorpAllMsg",
            "BountyClaimMsg",
            "BountyESSShared",
            "BountyESSTaken",
            "BountyPlacedAlliance",
            "BountyPlacedChar",
            "BountyPlacedCorp",
            "BountyYourBountyClaimed",
            "BuddyConnectContactAdd",
            "CharAppAcceptMsg",
            "CharAppRejectMsg",
            "CharAppWithdrawMsg",
            "CharLeftCorpMsg",
            "CharMedalMsg",
            "CharTerminationMsg",
            "CloneActivationMsg",
            "CloneActivationMsg2",
            "CloneMovedMsg",
            "CloneRevokedMsg1",
            "CloneRevokedMsg2",
            "CombatOperationFinished",
            "ContactAdd",
            "ContactEdit",
            "ContainerPasswordMsg",
            "ContractRegionChangedToPochven",
            "CorpAllBillMsg",
            "CorpAppAcceptMsg",
            "CorpAppInvitedMsg",
            "CorpAppNewMsg",
            "CorpAppRejectCustomMsg",
            "CorpAppRejectMsg",
            "CorpBecameWarEligible",
            "CorpDividendMsg",
            "CorpFriendlyFireDisableTimerCompleted",
            "CorpFriendlyFireDisableTimerStarted",
            "CorpFriendlyFireEnableTimerCompleted",
            "CorpFriendlyFireEnableTimerStarted",
            "CorpKicked",
            "CorpLiquidationMsg",
            "CorpNewCEOMsg",
            "CorpNewsMsg",
            "CorpNoLongerWarEligible",
            "CorpOfficeExpirationMsg",
            "CorpStructLostMsg",
            "CorpTaxChangeMsg",
            "CorpVoteCEORevokedMsg",
            "CorpVoteMsg",
            "CorpWarDeclaredMsg",
            "CorpWarDeclaredV2",
            "CorpWarFightingLegalMsg",
            "CorpWarInvalidatedMsg",
            "CorpWarRetractedMsg",
            "CorpWarSurrenderMsg",
            "CorporationGoalClosed",
            "CorporationGoalCompleted",
            "CorporationGoalCreated",
            "CorporationGoalNameChange",
            "CorporationLeft",
            "CustomsMsg",
            "DeclareWar",
            "DistrictAttacked",
            "DustAppAcceptedMsg",
            "ESSMainBankLink",
            "EntosisCaptureStarted",
            "ExpertSystemExpired",
            "ExpertSystemExpiryImminent",
            "FWAllianceKickMsg",
            "FWAllianceWarningMsg",
            "FWCharKickMsg",
            "FWCharRankGainMsg",
            "FWCharRankLossMsg",
            "FWCharWarningMsg",
            "FWCorpJoinMsg",
            "FWCorpKickMsg",
            "FWCorpLeaveMsg",
            "FWCorpWarningMsg",
            "FacWarCorpJoinRequestMsg",
            "FacWarCorpJoinWithdrawMsg",
            "FacWarCorpLeaveRequestMsg",
            "FacWarCorpLeaveWithdrawMsg",
            "FacWarLPDisqualifiedEvent",
            "FacWarLPDisqualifiedKill",
            "FacWarLPPayoutEvent",
            "FacWarLPPayoutKill",
            "GameTimeAdded",
            "GameTimeReceived",
            "GameTimeSent",
            "GiftReceived",
            "IHubDestroyedByBillFailure",
            "IncursionCompletedMsg",
            "IndustryOperationFinished",
            "IndustryTeamAuctionLost",
            "IndustryTeamAuctionWon",
            "InfrastructureHubBillAboutToExpire",
            "InsuranceExpirationMsg",
            "InsuranceFirstShipMsg",
            "InsuranceInvalidatedMsg",
            "InsuranceIssuedMsg",
            "InsurancePayoutMsg",
            "InvasionCompletedMsg",
            "InvasionSystemLogin",
            "InvasionSystemStart",
            "JumpCloneDeletedMsg1",
            "JumpCloneDeletedMsg2",
            "KillReportFinalBlow",
            "KillReportVictim",
            "KillRightAvailable",
            "KillRightAvailableOpen",
            "KillRightEarned",
            "KillRightUnavailable",
            "KillRightUnavailableOpen",
            "KillRightUsed",
            "LPAutoRedeemed",
            "LocateCharMsg",
            "MadeWarMutual",
            "MercOfferRetractedMsg",
            "MercOfferedNegotiationMsg",
            "MissionCanceledTriglavian",
            "MissionOfferExpirationMsg",
            "MissionTimeoutMsg",
            "MoonminingAutomaticFracture",
            "MoonminingExtractionCancelled",
            "MoonminingExtractionFinished",
            "MoonminingExtractionStarted",
            "MoonminingLaserFired",
            "MutualWarExpired",
            "MutualWarInviteAccepted",
            "MutualWarInviteRejected",
            "MutualWarInviteSent",
            "NPCStandingsGained",
            "NPCStandingsLost",
            "OfferToAllyRetracted",
            "OfferedSurrender",
            "OfferedToAlly",
            "OfficeLeaseCanceledInsufficientStandings",
            "OldLscMessages",
            "OperationFinished",
            "OrbitalAttacked",
            "OrbitalReinforced",
            "OwnershipTransferred",
            "RaffleCreated",
            "RaffleExpired",
            "RaffleFinished",
            "ReimbursementMsg",
            "ResearchMissionAvailableMsg",
            "RetractsWar",
            "SPAutoRedeemed",
            "SeasonalChallengeCompleted",
            "SkinSequencingCompleted",
            "SkyhookDeployed",
            "SkyhookDestroyed",
            "SkyhookLostShields",
            "SkyhookOnline",
            "SkyhookUnderAttack",
            "SovAllClaimAquiredMsg",
            "SovAllClaimLostMsg",
            "SovCommandNodeEventStarted",
            "SovCorpBillLateMsg",
            "SovCorpClaimFailMsg",
            "SovDisruptorMsg",
            "SovStationEnteredFreeport",
            "SovStructureDestroyed",
            "SovStructureReinforced",
            "SovStructureSelfDestructCancel",
            "SovStructureSelfDestructFinished",
            "SovStructureSelfDestructRequested",
            "SovereigntyIHDamageMsg",
            "SovereigntySBUDamageMsg",
            "SovereigntyTCUDamageMsg",
            "StationAggressionMsg1",
            "StationAggressionMsg2",
            "StationConquerMsg",
            "StationServiceDisabled",
            "StationServiceEnabled",
            "StationStateChangeMsg",
            "StoryLineMissionAvailableMsg",
            "StructureAnchoring",
            "StructureCourierContractChanged",
            "StructureDestroyed",
            "StructureFuelAlert",
            "StructureImpendingAbandonmentAssetsAtRisk",
            "StructureItemsDelivered",
            "StructureItemsMovedToSafety",
            "StructureLostArmor",
            "StructureLostShields",
            "StructureLowReagentsAlert",
            "StructureNoReagentsAlert",
            "StructureOnline",
            "StructurePaintPurchased",
            "StructureServicesOffline",
            "StructureUnanchoring",
            "StructureUnderAttack",
            "StructureWentHighPower",
            "StructureWentLowPower",
            "StructuresJobsCancelled",
            "StructuresJobsPaused",
            "StructuresReinforcementChanged",
            "TowerAlertMsg",
            "TowerResourceAlertMsg",
            "TransactionReversalMsg",
            "TutorialMsg",
            "WarAdopted",
            "WarAllyInherited",
            "WarAllyOfferDeclinedMsg",
            "WarConcordInvalidates",
            "WarDeclared",
            "WarEndedHqSecurityDrop",
            "WarHQRemovedFromSpace",
            "WarInherited",
            "WarInvalid",
            "WarRetracted",
            "WarRetractedByConcord",
            "WarSurrenderDeclinedMsg",
            "WarSurrenderOfferMsg",
        }
        # when
        for ntype in NotificationType.esi_notifications():
            with self.subTest(notification_type=ntype):
                self.assertIn(ntype, esi_valid_notification_types)



# Generated by Django 4.0.10 on 2023-04-17 00:47

import multiselectfield.db.fields

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("authentication", "0020_userprofile_language_userprofile_night_mode"),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("structures", "0002_remove_eveuniverse_relation_names"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="fuelalert",
            options={
                "verbose_name": "structure fuel alert",
                "verbose_name_plural": "structure fuel alerts",
            },
        ),
        migrations.AlterModelOptions(
            name="fuelalertconfig",
            options={
                "verbose_name": "structure fuel alert config",
                "verbose_name_plural": "structure fuel alert configs",
            },
        ),
        migrations.AlterModelOptions(
            name="generatednotification",
            options={
                "verbose_name": "generated notification",
                "verbose_name_plural": "generated  notifications",
            },
        ),
        migrations.AlterModelOptions(
            name="jumpfuelalert",
            options={
                "verbose_name": "jump fuel alert",
                "verbose_name_plural": "jump fuel alerts",
            },
        ),
        migrations.AlterModelOptions(
            name="jumpfuelalertconfig",
            options={
                "verbose_name": "jump fuel alert config",
                "verbose_name_plural": "jump fuel alert configs",
            },
        ),
        migrations.AlterModelOptions(
            name="notification",
            options={
                "verbose_name": "eve notification",
                "verbose_name_plural": "eve notifications",
            },
        ),
        migrations.AlterModelOptions(
            name="owner",
            options={"verbose_name": "owner", "verbose_name_plural": "owners"},
        ),
        migrations.AlterModelOptions(
            name="ownercharacter",
            options={
                "verbose_name": "owner character",
                "verbose_name_plural": "owner characters",
            },
        ),
        migrations.AlterModelOptions(
            name="structure",
            options={"verbose_name": "structure", "verbose_name_plural": "structures"},
        ),
        migrations.AlterModelOptions(
            name="structureitem",
            options={
                "verbose_name": "structure item",
                "verbose_name_plural": "structure items",
            },
        ),
        migrations.AlterModelOptions(
            name="structureservice",
            options={
                "verbose_name": "structure service",
                "verbose_name_plural": "structure services",
            },
        ),
        migrations.AlterModelOptions(
            name="structuretag",
            options={
                "ordering": ["order", "name"],
                "verbose_name": "structure tag",
                "verbose_name_plural": "structure tags",
            },
        ),
        migrations.AlterModelOptions(
            name="webhook",
            options={"verbose_name": "webhook", "verbose_name_plural": "webhooks"},
        ),
        migrations.AlterField(
            model_name="fuelalert",
            name="config",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="structure_fuel_alerts",
                to="structures.fuelalertconfig",
                verbose_name="configuration",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalert",
            name="hours",
            field=models.PositiveIntegerField(
                db_index=True,
                help_text="number of hours before fuel expiration this alert was sent",
                verbose_name="hours",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalert",
            name="structure",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="structure_fuel_alerts",
                to="structures.structure",
                verbose_name="structure",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalertconfig",
            name="color",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (14242639, "danger"),
                    (6013150, "info"),
                    (6076508, "success"),
                    (15773006, "warning"),
                ],
                default=15773006,
                help_text="Context color of these notification on Discord",
                null=True,
                verbose_name="color",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalertconfig",
            name="end",
            field=models.PositiveIntegerField(
                help_text="End of alerts in hours before fuel expires",
                verbose_name="end",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalertconfig",
            name="is_enabled",
            field=models.BooleanField(
                default=True,
                help_text="Disabled configurations will not create any new alerts.",
                verbose_name="is_enabled",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalertconfig",
            name="repeat",
            field=models.PositiveIntegerField(
                help_text="Notifications will be repeated every x hours. Set to 0 for no repeats",
                verbose_name="repeat",
            ),
        ),
        migrations.AlterField(
            model_name="fuelalertconfig",
            name="start",
            field=models.PositiveIntegerField(
                help_text="Start of alerts in hours before fuel expires",
                verbose_name="start",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="details",
            field=models.JSONField(default=dict, verbose_name="details"),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="is_sent",
            field=models.BooleanField(
                default=False,
                help_text="True when this notification has been forwarded to Discord",
                verbose_name="is sent",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="is_timer_added",
            field=models.BooleanField(
                default=False,
                help_text="True when a timer has been added for this notification",
                null=True,
                verbose_name="is timer added",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="last_updated",
            field=models.DateTimeField(auto_now=True, verbose_name="last_updated"),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="owner",
            field=models.ForeignKey(
                help_text="Corporation that owns this notification",
                on_delete=django.db.models.deletion.CASCADE,
                to="structures.owner",
                verbose_name="owner",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="structures",
            field=models.ManyToManyField(
                help_text="Structures this notification is about (if any)",
                to="structures.structure",
                verbose_name="structures",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="timestamp",
            field=models.DateTimeField(auto_now_add=True, verbose_name="timestamp"),
        ),
        migrations.AlterField(
            model_name="jumpfuelalert",
            name="config",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="jump_fuel_alerts",
                to="structures.jumpfuelalertconfig",
                verbose_name="configuration",
            ),
        ),
        migrations.AlterField(
            model_name="jumpfuelalert",
            name="structure",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="jump_fuel_alerts",
                to="structures.structure",
                verbose_name="structure",
            ),
        ),
        migrations.AlterField(
            model_name="jumpfuelalertconfig",
            name="color",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (14242639, "danger"),
                    (6013150, "info"),
                    (6076508, "success"),
                    (15773006, "warning"),
                ],
                default=15773006,
                help_text="Context color of these notification on Discord",
                null=True,
                verbose_name="color",
            ),
        ),
        migrations.AlterField(
            model_name="jumpfuelalertconfig",
            name="is_enabled",
            field=models.BooleanField(
                default=True,
                help_text="Disabled configurations will not create any new alerts.",
                verbose_name="is_enabled",
            ),
        ),
        migrations.AlterField(
            model_name="jumpfuelalertconfig",
            name="threshold",
            field=models.PositiveIntegerField(
                help_text="Notifications will be sent once fuel level in units reaches this threshold",
                verbose_name="threshold",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="created",
            field=models.DateTimeField(
                default=None,
                help_text="Date when this notification was first received from ESI",
                null=True,
                verbose_name="created",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="is_read",
            field=models.BooleanField(
                default=None,
                help_text="True when this notification has read in the eve client",
                null=True,
                verbose_name="is read",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="is_sent",
            field=models.BooleanField(
                default=False,
                help_text="True when this notification has been forwarded to Discord",
                verbose_name="is sent",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="is_timer_added",
            field=models.BooleanField(
                default=False,
                help_text="True when a timer has been added for this notification",
                null=True,
                verbose_name="is timer added",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="last_updated",
            field=models.DateTimeField(
                help_text="Date when this notification has last been updated from ESI",
                verbose_name="last updated",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="owner",
            field=models.ForeignKey(
                help_text="Corporation that owns this notification",
                on_delete=django.db.models.deletion.CASCADE,
                to="structures.owner",
                verbose_name="owner",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="sender",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="eveuniverse.eveentity",
                verbose_name="sender",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="structures",
            field=models.ManyToManyField(
                help_text="Structures this notification is about (if any)",
                to="structures.structure",
                verbose_name="structures",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="text",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="Notification details in YAML",
                null=True,
                verbose_name="text",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="timestamp",
            field=models.DateTimeField(db_index=True, verbose_name="timestamp"),
        ),
        migrations.AlterField(
            model_name="owner",
            name="are_pocos_public",
            field=models.BooleanField(
                default=False,
                help_text="whether pocos of this owner are shown on public POCO page",
                verbose_name="are pocos public",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="assets_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
                verbose_name="assets last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="corporation",
            field=models.OneToOneField(
                help_text="Corporation owning structures",
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="structure_owner",
                serialize=False,
                to="eveonline.evecorporationinfo",
                verbose_name="corporation",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="forwarding_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
                verbose_name="forwarding last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="has_default_pings_enabled",
            field=models.BooleanField(
                default=True,
                help_text="to enable or disable pinging of notifications for this owner e.g. with @everyone and @here",
                verbose_name="has default pings enabled",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="whether this owner is currently included in the sync process",
                verbose_name="is active",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_alliance_main",
            field=models.BooleanField(
                default=False,
                help_text="whether alliance wide notifications are forwarded for this owner (e.g. sov notifications)",
                verbose_name="is alliance main",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_included_in_service_status",
            field=models.BooleanField(
                default=True,
                help_text="whether the sync status of this owner is included in the overall status of this services",
                verbose_name="is included in service status",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_up",
            field=models.BooleanField(
                default=None,
                editable=False,
                help_text="whether all services for this owner are currently up",
                null=True,
                verbose_name="is up",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="notifications_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
                verbose_name="notifications last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification. ",
                related_name="+",
                to="auth.group",
                verbose_name="ping groups",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="structures_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
                verbose_name="structures last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Notifications are sent to these webhooks.",
                related_name="owners",
                to="structures.webhook",
                verbose_name="webhooks",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="character_ownership",
            field=models.ForeignKey(
                help_text="character used for syncing",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="authentication.characterownership",
                verbose_name="character_ownership",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="error_count",
            field=models.PositiveIntegerField(
                default=0,
                editable=False,
                help_text="Count of ESI errors which happened with this character.",
                verbose_name="error count",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="notifications_last_used_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                editable=False,
                help_text="when this character was last used for syncing notifications",
                null=True,
                verbose_name="notifications last used at",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="owner",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="characters",
                to="structures.owner",
                verbose_name="owner",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="structures_last_used_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                editable=False,
                help_text="when this character was last used for syncing structures",
                null=True,
                verbose_name="structures last used at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="created_at",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="date this structure was received from ESI for the first time",
                verbose_name="created at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_moon",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Moon next to this structure - if any",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.evemoon",
                verbose_name="moon",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_planet",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Planet next to this structure - if any",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.eveplanet",
                verbose_name="planet",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_solar_system",
            field=models.ForeignKey(
                help_text="Solar System the structure is located",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evesolarsystem",
                verbose_name="solar system",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_type",
            field=models.ForeignKey(
                help_text="Type of the structure",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evetype",
                verbose_name="type",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="fuel_expires_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date on which the structure will run out of fuel",
                null=True,
                verbose_name="fuel expires at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="has_core",
            field=models.BooleanField(
                blank=True,
                db_index=True,
                default=None,
                help_text="bool indicating if the structure has a quantum core",
                null=True,
                verbose_name="has core",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="has_fitting",
            field=models.BooleanField(
                blank=True,
                db_index=True,
                default=None,
                help_text="bool indicating if the structure has a fitting",
                null=True,
                verbose_name="has fitting",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="id",
            field=models.BigIntegerField(
                help_text="The Item ID of the structure",
                primary_key=True,
                serialize=False,
                verbose_name="id",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="last_online_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="date this structure had any of it's services online",
                null=True,
                verbose_name="last online at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="last_updated_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="date this structure was last updated from the EVE server",
                null=True,
                verbose_name="last updated at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="name",
            field=models.CharField(
                help_text="The full name of the structure",
                max_length=255,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="next_reinforce_apply",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                null=True,
                verbose_name="next reinforce apply",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="next_reinforce_hour",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                null=True,
                validators=[django.core.validators.MaxValueValidator(23)],
                verbose_name="next reinforce hour",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="owner",
            field=models.ForeignKey(
                help_text="Corporation that owns the structure",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="structures",
                to="structures.owner",
                verbose_name="owner",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_x",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="x position in the solar system",
                null=True,
                verbose_name="position x",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_y",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="y position in the solar system",
                null=True,
                verbose_name="position y",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_z",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="z position in the solar system",
                null=True,
                verbose_name="position z",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="reinforce_hour",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                help_text="The average hour of day that determines the time +/- some hours when the structure will randomly exit its reinforcement periods and become vulnerable to attack against its armor and/or hull. ",
                null=True,
                validators=[django.core.validators.MaxValueValidator(23)],
                verbose_name="reinforce hour",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="state",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "anchor vulnerable"),
                    (2, "anchoring"),
                    (3, "armor reinforce"),
                    (4, "armor vulnerable"),
                    (5, "deploy vulnerable"),
                    (6, "fitting invulnerable"),
                    (7, "hull reinforce"),
                    (8, "hull vulnerable"),
                    (9, "online deprecated"),
                    (10, "onlining vulnerable"),
                    (11, "shield vulnerable"),
                    (12, "unanchored"),
                    (21, "offline"),
                    (22, "online"),
                    (23, "onlining"),
                    (24, "reinforced"),
                    (25, "unanchoring "),
                    (0, "N/A"),
                    (13, "unknown"),
                ],
                default=13,
                help_text="Current state of the structure",
                verbose_name="state",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="state_timer_end",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date at which the structure entered it's current state",
                null=True,
                verbose_name="state timer end",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="state_timer_start",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date at which the structure will move to it's next state",
                null=True,
                verbose_name="state timer start",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="List of tags for this structure. ",
                related_name="structures",
                to="structures.structuretag",
                verbose_name="tags",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="unanchors_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date at which the structure will unanchor",
                null=True,
                verbose_name="unanchors at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Webhooks for sending notifications to. If any webhook is enabled, these will be used instead of the webhooks defined for the respective owner. If no webhook is enabled the owner's setting will be used. ",
                related_name="structures",
                to="structures.webhook",
                verbose_name="webhooks",
            ),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="eve_type",
            field=models.ForeignKey(
                help_text="type of the item",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evetype",
                verbose_name="type",
            ),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="id",
            field=models.BigIntegerField(
                help_text="The Eve item ID",
                primary_key=True,
                serialize=False,
                verbose_name="id",
            ),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="is_singleton",
            field=models.BooleanField(verbose_name="is singleton"),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="last_updated_at",
            field=models.DateTimeField(auto_now=True, verbose_name="last updated at"),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="location_flag",
            field=models.CharField(max_length=255, verbose_name="location flag"),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="quantity",
            field=models.IntegerField(verbose_name="quantity"),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="structure",
            field=models.ForeignKey(
                help_text="Structure this item is located in",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="items",
                to="structures.structure",
                verbose_name="structure",
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="name",
            field=models.CharField(
                help_text="Name of the service", max_length=100, verbose_name="name"
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="state",
            field=models.IntegerField(
                choices=[(1, "offline"), (2, "online")],
                help_text="Current state of this service",
                verbose_name="state",
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="structure",
            field=models.ForeignKey(
                help_text="Structure this service is installed to",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="services",
                to="structures.structure",
                verbose_name="structure",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="description",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="description for this tag",
                null=True,
                verbose_name="description",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="if true this custom tag will automatically be added to new structures",
                verbose_name="is default",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="is_user_managed",
            field=models.BooleanField(
                default=True,
                help_text="if False this tag is created and managed by the system and can not be modified by users",
                verbose_name="is user managed",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="name",
            field=models.CharField(
                help_text="name of the tag - must be unique",
                max_length=255,
                unique=True,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="order",
            field=models.PositiveIntegerField(
                blank=True,
                default=100,
                help_text="number defining the order tags are shown. custom tags can not have an order below 100",
                validators=[django.core.validators.MinValueValidator(100)],
                verbose_name="order",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="style",
            field=models.CharField(
                blank=True,
                choices=[
                    ("default", "grey"),
                    ("primary", "dark blue"),
                    ("success", "green"),
                    ("info", "light blue"),
                    ("warning", "orange"),
                    ("danger", "red"),
                ],
                default="default",
                help_text="color style of tag",
                max_length=16,
                verbose_name="style",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="has_default_pings_enabled",
            field=models.BooleanField(
                default=True,
                help_text="To enable or disable pinging of notifications for this webhook e.g. with @everyone and @here",
                verbose_name="has default pings enabled",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="Whether owners have this webhook automatically pre-set when created",
                verbose_name="is default",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="notification_types",
            field=multiselectfield.db.fields.MultiSelectField(
                choices=[
                    ("StructureAnchoring", "Upwell structure anchoring"),
                    ("StructureOnline", "Upwell structure went online"),
                    (
                        "StructureServicesOffline",
                        "Upwell structure services went offline",
                    ),
                    ("StructureWentHighPower", "Upwell structure went high power"),
                    ("StructureWentLowPower", "Upwell structure went low power"),
                    ("StructureUnanchoring", "Upwell structure unanchoring"),
                    ("StructureFuelAlert", "Upwell structure fuel alert"),
                    ("StructureRefueledExtra", "Upwell structure refueled"),
                    ("StructureJumpFuelAlert", "Upwell structure jump fuel alert"),
                    ("StructureUnderAttack", "Upwell structure is under attack"),
                    ("StructureLostShields", "Upwell structure lost shields"),
                    ("StructureLostArmor", "Upwell structure lost armor"),
                    ("StructureDestroyed", "Upwell structure destroyed"),
                    (
                        "StructuresReinforcementChanged",
                        "Upwell structure reinforcement time changed",
                    ),
                    ("OwnershipTransferred", "Upwell structure ownership transferred"),
                    ("OrbitalAttacked", "Customs office attacked"),
                    ("OrbitalReinforced", "Customs office reinforced"),
                    ("TowerAlertMsg", "Starbase attacked"),
                    ("TowerResourceAlertMsg", "Starbase fuel alert"),
                    ("TowerRefueledExtra", "Starbase refueled (BETA)"),
                    ("TowerReinforcedExtra", "Starbase reinforced (BETA)"),
                    ("MoonminingExtractionStarted", "Moon mining extraction started"),
                    ("MoonminingLaserFired", "Moonmining laser fired"),
                    (
                        "MoonminingExtractionCancelled",
                        "Moon mining extraction cancelled",
                    ),
                    ("MoonminingExtractionFinished", "Moon mining extraction finished"),
                    (
                        "MoonminingAutomaticFracture",
                        "Moon mining automatic fracture triggered",
                    ),
                    ("SovStructureReinforced", "Sovereignty structure reinforced"),
                    ("SovStructureDestroyed", "Sovereignty structure destroyed"),
                    ("EntosisCaptureStarted", "Sovereignty entosis capture started"),
                    (
                        "SovCommandNodeEventStarted",
                        "Sovereignty command node event started",
                    ),
                    ("SovAllClaimAquiredMsg", "Sovereignty claim acknowledgment"),
                    ("SovAllClaimLostMsg", "Sovereignty lost"),
                    ("AllAnchoringMsg", "Structure anchoring in alliance space"),
                    ("WarDeclared", "War declared"),
                    ("AllyJoinedWarAggressorMsg", "War ally joined aggressor"),
                    ("AllyJoinedWarAllyMsg", "War ally joined ally"),
                    ("AllyJoinedWarDefenderMsg", "War ally joined defender"),
                    ("WarAdopted", "War adopted"),
                    ("WarInherited", "War inherited"),
                    ("CorpWarSurrenderMsg", "War party surrendered"),
                    ("WarRetractedByConcord", "War retracted by Concord"),
                    ("CorpBecameWarEligible", "War corporation became eligible"),
                    ("CorpNoLongerWarEligible", "War corporation no longer eligible"),
                    ("WarSurrenderOfferMsg", "War surrender offered"),
                    ("CorpAppNewMsg", "Character submitted application"),
                    ("CorpAppInvitedMsg", "Character invited to join corporation"),
                    ("CorpAppRejectCustomMsg", "Corp application rejected"),
                    ("CharAppWithdrawMsg", "Character withdrew application"),
                    ("CharAppAcceptMsg", "Character joins corporation"),
                    ("CharLeftCorpMsg", "Character leaves corporation"),
                    ("BillOutOfMoneyMsg", "Bill out of money"),
                    (
                        "InfrastructureHubBillAboutToExpire",
                        "I-HUB bill about to expire",
                    ),
                    ("IHubDestroyedByBillFailure", "I_HUB destroyed by bill failure"),
                ],
                default=[
                    "StructureAnchoring",
                    "StructureDestroyed",
                    "StructureFuelAlert",
                    "StructureLostArmor",
                    "StructureLostShields",
                    "StructureOnline",
                    "StructureServicesOffline",
                    "StructureUnderAttack",
                    "StructureWentHighPower",
                    "StructureWentLowPower",
                    "OrbitalAttacked",
                    "OrbitalReinforced",
                    "TowerAlertMsg",
                    "TowerResourceAlertMsg",
                    "SovStructureReinforced",
                    "SovStructureDestroyed",
                ],
                help_text="Select which type of notifications should be forwarded to this webhook",
                max_length=1123,
                verbose_name="notification types",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification - ",
                related_name="+",
                to="auth.group",
                verbose_name="ping groups",
            ),
        ),
        migrations.AddConstraint(
            model_name="starbasedetailfuel",
            constraint=models.UniqueConstraint(
                fields=("detail", "eve_type"), name="functional_pk_starbasedetailfuel"
            ),
        ),
    ]



"""Models for webhooks."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .core import DiscordWebhookMixin
from .managers import WebhookBaseManager


# TODO: Investigate if this can be changed to a dependency injection to use fakes
class WebhookBase(DiscordWebhookMixin, models.Model):
    """Base model for a Webhook"""

    class PingType(models.TextChoices):
        """A ping type for webhooks."""

        NONE = "NO", _("none")
        HERE = "HE", _("here")
        EVERYONE = "EV", _("everyone")

    class Color(models.IntegerChoices):
        """A color for embeds."""

        DANGER = 0xD9534F, _("danger")
        INFO = 0x5BC0DE, _("info")
        SUCCESS = 0x5CB85C, _("success")
        WARNING = 0xF0AD4E, _("warning")

        @property
        def css_color(self) -> str:
            """Return color as CSS value."""
            return f"#{self.value:X}"

    TYPE_DISCORD = 1

    TYPE_CHOICES = [
        (TYPE_DISCORD, _("Discord Webhook")),
    ]

    name = models.CharField(
        verbose_name=_("name"),
        max_length=64,
        unique=True,
        help_text=_("Short name to identify this webhook"),
    )
    url = models.CharField(
        verbose_name=_("url"),
        max_length=255,
        unique=True,
        help_text=_(
            "URL of this webhook, e.g. "
            "https://discordapp.com/api/webhooks/123456/abcdef"
        ),
    )
    notes = models.TextField(
        verbose_name=_("notes"),
        null=True,
        default=None,
        blank=True,
        help_text=_("Notes regarding this webhook"),
    )
    webhook_type = models.IntegerField(
        verbose_name=_("webhook type"),
        choices=TYPE_CHOICES,
        default=TYPE_DISCORD,
        help_text=_("Type of this webhook"),
    )
    is_active = models.BooleanField(
        verbose_name=_("is active"),
        default=True,
        help_text=_("Whether notifications are currently sent to this webhook"),
    )

    objects = WebhookBaseManager()

    class Meta:
        abstract = True



"""Starbase embeds."""

# pylint: disable=missing-class-docstring

import datetime as dt

import dhooks_lite

from django.utils.translation import gettext as _

from structures.core import starbases
from structures.models import GeneratedNotification, Notification, Structure, Webhook

from .helpers import (
    gen_corporation_link,
    gen_solar_system_text,
    target_datetime_formatted,
    timeuntil,
)
from .main import NotificationBaseEmbed


class NotificationTowerEmbed(NotificationBaseEmbed):
    """Base class for all tower (aka POS) related notification embeds."""

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        eve_moon = self._notification.eve_moon()
        structure_type = self._notification.eve_structure_type("typeID")
        self._structure = Structure.objects.filter(eve_moon=eve_moon).first()
        if self._structure:
            structure_name = self._structure.name
        else:
            structure_name = structure_type.name

        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._description = (
            _(
                "The starbase %(structure_name)s at %(moon)s "
                "in %(solar_system)s belonging to %(owner_link)s"
            )
            % {
                "structure_name": Webhook.text_bold(structure_name),
                "moon": eve_moon.name,
                "solar_system": gen_solar_system_text(
                    eve_moon.eve_planet.eve_solar_system
                ),
                "owner_link": gen_corporation_link(str(notification.owner)),
            }
            + " "
        )


class NotificationTowerAlertMsg(NotificationTowerEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        aggressor_link = self.gen_aggressor_link()
        damage_text = self.compile_damage_text("Value", 100)
        self._title = _("Starbase under attack")
        self._description += _("is under attack by %(aggressor)s.\n%(damage_text)s") % {
            "aggressor": aggressor_link,
            "damage_text": damage_text,
        }
        self._color = Webhook.Color.WARNING


class NotificationTowerResourceAlertMsg(NotificationTowerEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        if "wants" in self._data and self._data["wants"]:
            fuel_quantity = self._data["wants"][0]["quantity"]
            starbase_type = self._notification.eve_structure_type("typeID")
            seconds = starbases.fuel_duration(
                starbase_type=starbase_type,
                fuel_quantity=fuel_quantity,
                has_sov=notification.owner.has_sov(
                    self._notification.eve_solar_system()
                ),
            )
            from_date = self.notification.timestamp
            to_date = from_date + dt.timedelta(seconds=seconds)
            hours_left = timeuntil(to_date, from_date)
        elif self._structure and self._structure.fuel_expires_at:
            hours_left = timeuntil(self._structure.fuel_expires_at)
        else:
            hours_left = "?"
        self._title = _("Starbase fuel alert")
        self._description += _("is running out of fuel in %s.") % Webhook.text_bold(
            hours_left
        )
        self._color = Webhook.Color.WARNING


class NotificationTowerRefueledExtra(NotificationTowerEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        target_date = self.fuel_expires_target_date()
        self._title = _("Starbase refueled")
        self._description += (
            _("has been refueled. Fuel will last until %s.") % target_date
        )
        self._color = Webhook.Color.INFO


class GeneratedNotificationBaseEmbed(NotificationBaseEmbed):
    """Base class for generated notification embeds."""

    def __init__(self, notification: GeneratedNotification) -> None:
        super().__init__(notification)
        if not isinstance(notification, GeneratedNotification):
            raise TypeError(
                "Can only create embeds from GeneratedNotification objects."
            )
        self._structure = notification.structures.first()
        self._thumbnail = dhooks_lite.Thumbnail(
            self._structure.eve_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class GeneratedNotificationTowerEmbed(GeneratedNotificationBaseEmbed):
    """Base class for all tower (aka POS) related generated notification embeds."""

    def __init__(self, notification: GeneratedNotification) -> None:
        super().__init__(notification)
        self._description = _(
            "The starbase %(structure_name)s at %(moon)s "
            "in %(solar_system)s belonging to %(owner_link)s "
        ) % {
            "structure_name": Webhook.text_bold(self._structure.name),
            "moon": self._structure.eve_moon.name,
            "solar_system": gen_solar_system_text(self._structure.eve_solar_system),
            "owner_link": gen_corporation_link(str(notification.owner)),
        }


class NotificationTowerReinforcedExtra(GeneratedNotificationTowerEmbed):
    def __init__(self, notification: GeneratedNotification) -> None:
        super().__init__(notification)
        self._title = _("Starbase reinforced")
        try:
            reinforced_until = target_datetime_formatted(
                dt.datetime.fromisoformat(notification.details["reinforced_until"])
            )
        except (KeyError, ValueError):
            reinforced_until = _("(unknown)")

        self._description += (
            _("has been reinforced and will come out at: %s.") % reinforced_until
        )

        self._color = Webhook.Color.DANGER



"""Poco embeds."""

# pylint: disable=missing-class-docstring

import dhooks_lite

from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime

from structures.constants import EveTypeId
from structures.helpers import get_or_create_eve_type
from structures.models import Notification, Webhook

from .helpers import (
    gen_corporation_link,
    gen_solar_system_text,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class NotificationOrbitalEmbed(NotificationBaseEmbed):
    """Base class for all orbital (aka POCO) related notification embeds."""

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._planet = self._notification.eve_planet()
        self._structure_type = get_or_create_eve_type(id=EveTypeId.CUSTOMS_OFFICE)
        self._solar_system_link = gen_solar_system_text(
            self._notification.eve_solar_system()
        )
        self._owner_link = gen_corporation_link(str(notification.owner))
        self._aggressor_link = self.gen_aggressor_link()
        self._thumbnail = dhooks_lite.Thumbnail(
            self._structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationOrbitalAttacked(NotificationOrbitalEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Orbital under attack")
        self._description = _(
            "The %(structure_type)s at %(planet)s in %(solar_system)s "
            "belonging to %(owner_link)s "
            "is under attack by %(aggressor)s."
        ) % {
            "structure_type": self._structure_type.name,
            "planet": self._planet.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "aggressor": self._aggressor_link,
        }
        self._color = Webhook.Color.WARNING


class NotificationOrbitalReinforced(NotificationOrbitalEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        reinforce_exit_time = ldap_time_2_datetime(self._data["reinforceExitTime"])
        self._title = _("Orbital reinforced")
        self._description = _(
            "The %(structure_type)s at %(planet)s in %(solar_system)s "
            "belonging to %(owner_link)s "
            "has been reinforced by %(aggressor)s "
            "and will come out at: %(date)s."
        ) % {
            "structure_type": self._structure_type.name,
            "planet": self._planet.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "aggressor": self._aggressor_link,
            "date": target_datetime_formatted(reinforce_exit_time),
        }
        self._color = Webhook.Color.DANGER



import datetime as dt

from django.test import TestCase
from eveuniverse.models import EveEntity

from structures.core.notification_embeds import helpers
from structures.tests.testdata.factories import EveEntityCharacterFactory


class TestTimeuntil(TestCase):
    def test_should_return_time(self):
        # given
        to_date = dt.datetime(2023, 7, 29, 12, 00, tzinfo=dt.timezone.utc)
        from_date = dt.datetime(2023, 7, 28, 11, 00, tzinfo=dt.timezone.utc)
        # when
        result = helpers.timeuntil(to_date=to_date, from_date=from_date)
        # then
        self.assertEqual("a day from now", result)


class TestGenEveEntityLinkFromId(TestCase):
    def test_should_return_eve_entity(self):
        # given
        obj = EveEntityCharacterFactory()
        # when
        result = helpers.gen_eve_entity_link_from_id(obj.id)
        # then
        self.assertIn(str(obj.id), result)

    def test_should_return_empty_string(self):
        self.assertEqual(helpers.gen_eve_entity_link_from_id(None), "")


class TestGenEveEntityLink(TestCase):
    def test_should_return_eve_entity(self):
        # given
        obj = EveEntityCharacterFactory()
        # when
        result = helpers.gen_eve_entity_link(obj)
        # then
        self.assertIn(str(obj.id), result)

    def test_should_return_empty_string_when_obj_not_valid(self):
        # when
        result = helpers.gen_eve_entity_link(None)
        # then
        self.assertEqual(result, "")

    def test_should_return_empty_string_when_eve_entity_not_supported(self):
        # given
        obj = EveEntity(id=99, name="special", category=EveEntity.CATEGORY_SOLAR_SYSTEM)
        # when
        result = helpers.gen_eve_entity_link(obj)
        # then
        self.assertEqual(result, "")



from typing import List

from django.test import RequestFactory

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
)
from app_utils.testing import NoSocketsTestCase

from structures.core.serializers import PocoListSerializer, StructureListSerializer
from structures.models import Structure
from structures.tests.testdata.factories import (
    JumpGateFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    UserMainDefaultFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse


def to_dict(lst: List[dict], key="id"):
    return {obj[key]: obj for obj in lst}


class TestStructureListSerializer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainDefaultFactory()
        cls.owner = OwnerFactory(user=cls.user)
        alliance = EveAllianceInfoFactory(
            alliance_name="Wayne Enterprises", alliance_ticker="WYE"
        )
        corporation = EveCorporationInfoFactory(
            corporation_name="Wayne Technologies", alliance=alliance
        )
        character = EveCharacterFactory(corporation=corporation)
        cls.user = UserMainDefaultFactory(main_character__character=character)
        cls.owner = OwnerFactory(corporation=corporation)

    def _render_structure(self, structure: Structure, queryset=None) -> dict:
        """Return rendered data row for a structure."""
        request = self.factory.get("/")
        request.user = self.user
        my_queryset = queryset or Structure.objects.all()
        data = StructureListSerializer(queryset=my_queryset, request=request).to_list()
        obj = to_dict(data)[structure.id]
        return obj

    def test_should_render_upwell_structure(self):
        # given
        structure = StructureFactory(
            owner=self.owner, eve_solar_system_name="Amamake", eve_type_name="Astrahus"
        )
        # when
        obj = self._render_structure(structure)
        # then
        self.assertEqual(obj["alliance_name"], "Wayne Enterprises [WYE]")
        self.assertEqual(obj["corporation_name"], "Wayne Technologies")
        self.assertEqual(obj["region_name"], "Heimatar")
        self.assertEqual(obj["solar_system_name"], "Amamake")
        self.assertEqual(obj["group_name"], "Citadel")
        self.assertEqual(obj["category_name"], "Structure")
        self.assertFalse(obj["is_starbase"])
        self.assertFalse(obj["is_poco"])
        self.assertEqual(obj["type_name"], "Astrahus")
        self.assertFalse(obj["is_reinforced"])
        self.assertEqual(obj["is_reinforced_str"], "no")
        self.assertEqual(
            obj["fuel_and_power"]["fuel_expires_at"],
            structure.fuel_expires_at.isoformat(),
        )
        self.assertEqual(obj["power_mode_str"], "Full Power")
        self.assertEqual(obj["state_str"], "Shield vulnerable")
        self.assertEqual(obj["core_status_str"], "yes")
        self.assertEqual(obj["details"], "")

    def test_should_show_reinforced_for_structure(self):
        # given
        structure = StructureFactory(
            owner=self.owner, state=Structure.State.ARMOR_REINFORCE
        )
        # when
        obj = self._render_structure(structure)
        # then
        self.assertTrue(obj["is_reinforced"])

    def test_should_show_not_reinforced_for_starbase(self):
        # given
        structure = StarbaseFactory(owner=self.owner)
        # when
        obj = self._render_structure(structure)
        # then
        self.assertFalse(obj["is_reinforced"])

    def test_should_show_reinforced_for_starbase(self):
        # given
        structure = StarbaseFactory(
            owner=self.owner, state=Structure.State.POS_REINFORCED
        )
        # when
        obj = self._render_structure(structure)
        # then
        self.assertTrue(obj["is_reinforced"])

    def test_should_handle_owner_without_alliance(self):
        # given
        corporation = EveCorporationInfoFactory(create_alliance=False)
        owner = OwnerFactory(corporation=corporation)
        structure = StarbaseFactory(owner=owner)
        # when
        obj = self._render_structure(structure)
        # then
        self.assertEqual(obj["alliance_name"], "")

    def test_should_return_jump_gates(self):
        # given
        structure = JumpGateFactory(
            owner=self.owner, jump_fuel_quantity=5000, eve_solar_system_name="1-PGSG"
        )
        # when
        obj = self._render_structure(
            structure, queryset=Structure.objects.annotate_jump_fuel_quantity()
        )
        # then
        self.assertEqual(obj["region_name"], "Detorid")
        self.assertEqual(obj["solar_system_name"], "1-PGSG")
        # self.assertEqual(
        #     obj["structure_name_and_tags"], "1-PGSG &gt;&gt; A-C5TC - Test Jump Gate"
        # )
        self.assertEqual(obj["jump_fuel_quantity"], 5000)


class TestPocoListSerializer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        alliance = EveAllianceInfoFactory(
            alliance_name="Wayne Enterprises", alliance_ticker="WYE"
        )
        corporation = EveCorporationInfoFactory(
            corporation_name="Wayne Technologies", alliance=alliance
        )
        cls.character = EveCharacterFactory(corporation=corporation)
        cls.user = UserMainDefaultFactory(main_character__character=cls.character)
        cls.owner = OwnerFactory(user=cls.user, are_pocos_public=True)

    def _render_structure(self, structure: Structure, queryset=None) -> dict:
        """Return rendered data row for a structure."""
        request = self.factory.get("/")
        request.user = self.user
        my_queryset = queryset or Structure.objects.all()
        data = PocoListSerializer(
            queryset=my_queryset, request=request, character=self.character
        ).to_list()
        obj = to_dict(data)[structure.id]
        return obj

    def test_should_render_poco_correctly(self):
        # given
        structure = PocoFactory(
            owner=self.owner,
            eve_planet_name="Amamake V",
            poco_details__corporation_tax_rate=0.01,
        )
        # when
        obj = self._render_structure(structure)
        # then
        self.assertEqual(obj["alliance_name"], "Wayne Enterprises [WYE]")
        self.assertIn("Hed", obj["constellation"])
        self.assertEqual(obj["corporation_name"], "Wayne Technologies")
        self.assertEqual(obj["has_access_str"], "yes")
        self.assertEqual(obj["region"], "Heimatar")
        self.assertEqual(obj["planet_name"], "Amamake V")
        self.assertEqual(obj["planet_type_name"], "Barren")
        self.assertEqual(obj["solar_system"], "Amamake")
        self.assertEqual(obj["space_type"], "lowsec")
        self.assertEqual(obj["tax"]["sort"], 1.0)



# Generated by Django 4.0.10 on 2024-06-23 14:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0005_add_notification_types"),
    ]

    operations = [
        migrations.AddField(
            model_name="ownercharacter",
            name="disabled_reason",
            field=models.TextField(default=""),
        ),
        migrations.AddField(
            model_name="ownercharacter",
            name="is_enabled",
            field=models.BooleanField(
                default=True,
                help_text="Disabled characters are not used for syncing owners",
                verbose_name="is enabled",
            ),
        ),
    ]



"""Models related to Structure."""

import datetime as dt
from typing import NamedTuple, Optional

from django.db import models
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from structures import __title__
from structures.constants import EveGroupId
from structures.core import starbases

from .structures_1 import Structure

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class StructureService(models.Model):
    """Service of a Structure."""

    class State(models.IntegerChoices):
        """A state of a structure service."""

        OFFLINE = 1, _("offline")
        ONLINE = 2, _("online")

        @classmethod
        def from_esi_name(cls, esi_state_name: str) -> "StructureService.State":
            """Return matching state for given ESI state name."""
            states_esi_map = {"offline": cls.OFFLINE, "online": cls.ONLINE}
            return (
                states_esi_map[esi_state_name]
                if esi_state_name in states_esi_map
                else cls.OFFLINE
            )

    structure = models.ForeignKey(
        Structure,
        on_delete=models.CASCADE,
        related_name="services",
        verbose_name=_("structure"),
        help_text=_("Structure this service is installed to"),
    )
    name = models.CharField(
        max_length=100, verbose_name=_("name"), help_text=_("Name of the service")
    )

    state = models.IntegerField(
        choices=State.choices,
        verbose_name=_("state"),
        help_text=_("Current state of this service"),
    )

    class Meta:
        verbose_name = _("structure service")
        verbose_name_plural = _("structure services")
        unique_together = (("structure", "name"),)

    def __str__(self):
        return f"{self.structure} - {self.name}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"structure_id={self.structure.id}, name='{self.name}')"
        )


class PocoDetails(models.Model):
    """Additional information about a POCO."""

    class StandingLevel(models.IntegerChoices):
        """A standing level in Eve Online."""

        NONE = -99, _("none")
        TERRIBLE = -10, _("terrible")
        BAD = -5, _("bad")
        NEUTRAL = 0, _("neutral")
        GOOD = 5, _("good")
        EXCELLENT = 10, _("excellent")

        @classmethod
        def from_esi(cls, value) -> "PocoDetails.StandingLevel":
            """Return match from ESI value or NONE"""
            my_map = {
                "bad": cls.BAD,
                "excellent": cls.EXCELLENT,
                "good": cls.GOOD,
                "neutral": cls.NEUTRAL,
                "terrible": cls.TERRIBLE,
            }
            try:
                return my_map[value]
            except KeyError:
                return cls.NONE

    class PocoCharacterAccessInfo(NamedTuple):
        """Access information to a poco for a character."""

        character_id: int
        has_access: bool
        is_confident: bool
        tax_rate: float

    structure = models.OneToOneField(
        Structure, on_delete=models.CASCADE, related_name="poco_details"
    )

    alliance_tax_rate = models.FloatField(null=True, default=None)
    allow_access_with_standings = models.BooleanField()
    allow_alliance_access = models.BooleanField()
    bad_standing_tax_rate = models.FloatField(null=True, default=None)
    corporation_tax_rate = models.FloatField(null=True, default=None)
    excellent_standing_tax_rate = models.FloatField(null=True, default=None)
    good_standing_tax_rate = models.FloatField(null=True, default=None)
    neutral_standing_tax_rate = models.FloatField(null=True, default=None)
    reinforce_exit_end = models.PositiveIntegerField()
    reinforce_exit_start = models.PositiveIntegerField()
    standing_level = models.IntegerField(
        choices=StandingLevel.choices, default=StandingLevel.NONE
    )
    terrible_standing_tax_rate = models.FloatField(null=True, default=None)

    def __str__(self):
        return f"{self.structure}"

    @property
    def reinforce_exit_end_str(self) -> str:
        """Return reinforce exit end as string."""
        return f"{self.reinforce_exit_end}:00"

    @property
    def reinforce_exit_start_str(self) -> str:
        """Return reinforce exit start as string."""
        return f"{self.reinforce_exit_start}:00"

    def determine_access_and_tax_for_character(
        self, character: EveCharacter
    ) -> "PocoCharacterAccessInfo":
        """Return access and tax information for a character."""
        owner_corporation = self.structure.owner.corporation

        if character.corporation_id == owner_corporation.corporation_id:
            # Corporation member
            return self.PocoCharacterAccessInfo(
                character_id=character.character_id,
                has_access=True,
                is_confident=True,
                tax_rate=self.corporation_tax_rate,
            )

        if (
            character.alliance_id
            and owner_corporation.alliance
            and owner_corporation.alliance.alliance_id == character.alliance_id
            and self.allow_alliance_access
            and self.alliance_tax_rate is not None
        ):
            # Alliance member
            return self.PocoCharacterAccessInfo(
                character_id=character.character_id,
                has_access=True,
                is_confident=True,
                tax_rate=self.alliance_tax_rate,
            )

        # neutral character
        has_access = (
            self.allow_access_with_standings
            and self.standing_level <= self.StandingLevel.NEUTRAL
            and self.neutral_standing_tax_rate is not None
        )
        is_confident = False

        if has_access:
            tax_rate = self.neutral_standing_tax_rate
        else:
            tax_rate = None

        return self.PocoCharacterAccessInfo(
            character_id=character.character_id,
            has_access=has_access,
            is_confident=is_confident,
            tax_rate=tax_rate,
        )

    def standing_level_access_map(self) -> dict:
        """Return map of access per standing level with standing level names as key."""
        names_map = {
            self.StandingLevel.NONE: "NONE",
            self.StandingLevel.TERRIBLE: "TERRIBLE",
            self.StandingLevel.BAD: "BAD",
            self.StandingLevel.NEUTRAL: "NEUTRAL",
            self.StandingLevel.GOOD: "GOOD",
            self.StandingLevel.EXCELLENT: "EXCELLENT",
        }
        return {
            names_map[self.StandingLevel(level)]: (
                self.allow_access_with_standings and level >= self.standing_level
            )
            for level in self.StandingLevel.values
        }


class StarbaseDetail(models.Model):
    """Additional information about a starbase."""

    class Role(models.TextChoices):
        """A role for a starbase."""

        ALLIANCE_MEMBER = "AL", _("alliance member")
        CONFIG_STARBASE_EQUIPMENT_ROLE = "CE", _("config starbase equipment role")
        CORPORATION_MEMBER = "CO", _("corporation member")
        STARBASE_FUEL_TECHNICIAN_ROLE = "FT", _("starbase fuel technician role")

        @classmethod
        def from_esi(cls, name: str) -> "StarbaseDetail.Role":
            """Create role from ESI name."""
            my_map = {
                "alliance_member": cls.ALLIANCE_MEMBER,
                "config_starbase_equipment_role": cls.CONFIG_STARBASE_EQUIPMENT_ROLE,
                "corporation_member": cls.CORPORATION_MEMBER,
                "starbase_fuel_technician_role": cls.STARBASE_FUEL_TECHNICIAN_ROLE,
            }
            return my_map[name]

    structure = models.OneToOneField(
        Structure, on_delete=models.CASCADE, related_name="starbase_detail"
    )

    allow_alliance_members = models.BooleanField()
    allow_corporation_members = models.BooleanField()
    anchor_role = models.CharField(max_length=2, choices=Role.choices)
    attack_if_at_war = models.BooleanField()
    attack_if_other_security_status_dropping = models.BooleanField()
    attack_security_status_threshold = models.FloatField(default=None, null=True)
    attack_standing_threshold = models.FloatField(default=None, null=True)
    fuel_bay_take_role = models.CharField(max_length=2, choices=Role.choices)
    fuel_bay_view_role = models.CharField(max_length=2, choices=Role.choices)
    last_modified_at = models.DateTimeField(
        help_text="When data was modified on the server."
    )
    offline_role = models.CharField(max_length=2, choices=Role.choices)
    online_role = models.CharField(max_length=2, choices=Role.choices)
    unanchor_role = models.CharField(max_length=2, choices=Role.choices)
    use_alliance_standings = models.BooleanField()

    def __str__(self) -> str:
        return str(self.structure)

    def calc_fuel_expires(self) -> Optional[dt.datetime]:
        """Estimate when fuel will expire for this starbase.

        Estimate will vary due to server caching of remaining fuel blocks.
        """
        if self.structure.state is Structure.State.POS_OFFLINE:
            return None

        fuel = self.fuels.filter(eve_type__eve_group_id=EveGroupId.FUEL_BLOCK).first()
        if not fuel:
            return None

        seconds = starbases.fuel_duration(
            starbase_type=self.structure.eve_type,
            fuel_quantity=fuel.quantity,
            has_sov=self.structure.owner.has_sov(self.structure.eve_solar_system),
        )
        return self.last_modified_at + dt.timedelta(seconds=seconds)


class StarbaseDetailFuel(models.Model):
    """Fuel for a starbase detail."""

    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    detail = models.ForeignKey(
        StarbaseDetail, on_delete=models.CASCADE, related_name="fuels"
    )
    quantity = models.IntegerField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["detail", "eve_type"], name="functional_pk_starbasedetailfuel"
            )
        ]

    def __str__(self) -> str:
        return f"{self.detail}-{self.eve_type}"



# flake8: noqa
"""Script for creating generated notifications for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
import datetime as dt

from django.utils.timezone import now

from allianceauth.eveonline.models import EveCorporationInfo

from structures.models import Owner, Webhook  # noqa: E402, E501
from structures.tests.testdata.factories import GeneratedNotificationFactory

# corporation the notifications will be generated for
CORPORATION_ID = 1000127  # Guristas

try:
    corporation = EveCorporationInfo.objects.get(corporation_id=CORPORATION_ID)
except EveCorporationInfo.DoesNotExist:
    corporation = EveCorporationInfo.objects.create_corporation(CORPORATION_ID)
owner, created = Owner.objects.get_or_create(
    corporation=corporation, defaults={"is_active": False}
)
if created and not owner.webhooks.exists():
    webhook = Webhook.objects.filter(is_active=True, is_default=True).first()
    if webhook:
        owner.webhooks.add(webhook)

reinforced_until = now() + dt.timedelta(hours=24)
notif = GeneratedNotificationFactory(owner=owner)
print(f"Created new generated notification: {notif}")



"""Helpers for generating embeds."""

import datetime as dt
from typing import Optional

import humanize

from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveSolarSystem

from allianceauth.eveonline.evelinks import dotlan, evewho
from app_utils.datetime import DATETIME_FORMAT

from structures.models import Webhook


def convert_to_naive_datetime(value: dt.datetime) -> dt.datetime:
    """Convert aware datetime to naive datetime and passthrough any other type."""
    if value.tzinfo is not None:
        value = dt.datetime.fromtimestamp(value.timestamp())
    return value


def timeuntil(to_date: dt.datetime, from_date: Optional[dt.datetime] = None) -> str:
    """Render timeuntil template tag for given datetime to string."""
    if not from_date:
        from_date = now()
    to_date_naive = convert_to_naive_datetime(to_date)
    from_date_naive = convert_to_naive_datetime(from_date)
    return humanize.naturaltime(to_date_naive, when=from_date_naive)


def target_datetime_formatted(target_datetime: dt.datetime) -> str:
    """Formatted Discord string for a target datetime."""
    return (
        f"{Webhook.text_bold(target_datetime.strftime(DATETIME_FORMAT))} "
        f"({timeuntil(target_datetime)})"
    )


def gen_solar_system_text(solar_system: EveSolarSystem) -> str:
    """Return generated text for a solar system."""
    solar_system_link = Webhook.create_link(
        solar_system.name, dotlan.solar_system_url(solar_system.name)
    )
    region_name = solar_system.eve_constellation.eve_region.name
    text = f"{solar_system_link} ({region_name})"
    return text


def gen_alliance_link(alliance_name: str) -> str:
    """Return generated text for an alliance link."""
    return Webhook.create_link(alliance_name, dotlan.alliance_url(alliance_name))


def gen_eve_entity_link(eve_entity: EveEntity) -> str:
    """Return a link for an alliance, corporation or character from an eve entity.

    Or return an empty string for any other category.
    """
    try:
        category = eve_entity.category
    except AttributeError:
        return ""

    if category == EveEntity.CATEGORY_ALLIANCE:
        url = dotlan.alliance_url(eve_entity.name)

    elif category == EveEntity.CATEGORY_CORPORATION:
        url = dotlan.corporation_url(eve_entity.name)

    elif category == EveEntity.CATEGORY_CHARACTER:
        url = evewho.character_url(eve_entity.id)

    else:
        return ""

    return Webhook.create_link(eve_entity.name, url)


def gen_eve_entity_link_from_id(id: int) -> str:
    """Return a link for an alliance, corporation or character from an ID."""
    if not id:
        return ""
    entity, _ = EveEntity.objects.get_or_create_esi(id=id)
    return gen_eve_entity_link(entity)


def gen_corporation_link(corporation_name: str) -> str:
    """Return a link for a corporation."""
    return Webhook.create_link(
        corporation_name, dotlan.corporation_url(corporation_name)
    )



from eveuniverse.models import EveSolarSystem

from app_utils.testing import NoSocketsTestCase

from structures.models import EveSovereigntyMap, EveSpaceType
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.eveuniverse"


class TestEveSovereigntyMap(NoSocketsTestCase):
    def test_str(self):
        obj = EveSovereigntyMap(solar_system_id=99)
        expected = "99"
        self.assertEqual(str(obj), expected)

    def test_repr(self):
        obj = EveSovereigntyMap(solar_system_id=99)
        expected = "EveSovereigntyMap(solar_system_id='99')"
        self.assertEqual(repr(obj), expected)


class TestEveSpaceType(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_space_type_highsec(self):
        obj = EveSolarSystem.objects.get(id=30002506)
        expected = EveSpaceType.HIGHSEC
        self.assertEqual(EveSpaceType.from_solar_system(obj), expected)

    def test_space_type_lowsec(self):
        obj = EveSolarSystem.objects.get(id=30002537)
        expected = EveSpaceType.LOWSEC
        self.assertEqual(EveSpaceType.from_solar_system(obj), expected)

    def test_space_type_nullsec(self):
        obj = EveSolarSystem.objects.get(id=30000474)
        expected = EveSpaceType.NULLSEC
        self.assertEqual(EveSpaceType.from_solar_system(obj), expected)

    def test_space_type_wh_space(self):
        obj = EveSolarSystem.objects.get(id=31000005)
        expected = EveSpaceType.W_SPACE
        self.assertEqual(EveSpaceType.from_solar_system(obj), expected)



"""Moonmining embeds."""

# pylint: disable=missing-class-docstring


import dhooks_lite

from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime

from structures.app_settings import STRUCTURES_NOTIFICATION_SHOW_MOON_ORE
from structures.helpers import get_or_create_eve_entity, get_or_create_eve_type
from structures.models import Notification, Webhook

from .helpers import (
    gen_corporation_link,
    gen_solar_system_text,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class NotificationMoonminingEmbed(NotificationBaseEmbed):
    """Base class for all moon mining related notification embeds."""

    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._moon = self._notification.eve_moon()
        self._solar_system_link = gen_solar_system_text(
            self._notification.eve_solar_system()
        )
        self._structure_name = self._data["structureName"]
        self._owner_link = gen_corporation_link(str(notification.owner))
        structure_type = self._notification.eve_structure_type()
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self.ore_text = (
            _("Estimated ore composition: %s") % self._ore_composition_text()
            if STRUCTURES_NOTIFICATION_SHOW_MOON_ORE
            else ""
        )

    def _ore_composition_text(self) -> str:
        if "oreVolumeByType" not in self._data:
            return ""

        ore_list = []
        for ore_type_id, volume in self._data["oreVolumeByType"].items():
            ore_type = get_or_create_eve_type(id=ore_type_id)
            if ore_type:
                ore_list.append(
                    {"id": ore_type_id, "name": ore_type.name, "volume": volume}
                )

        ore_list_2 = sorted(ore_list, key=lambda x: x["name"])
        return "\n- " + "\n- ".join(
            [f"{ore['name']}: {ore['volume']:,.0f} m³" for ore in ore_list_2]
        )


class NotificationMoonminningExtractionStarted(NotificationMoonminingEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        started_by = get_or_create_eve_entity(id=self._data["startedBy"])
        ready_time = ldap_time_2_datetime(self._data["readyTime"])
        auto_time = ldap_time_2_datetime(self._data["autoTime"])
        self._title = _("Moon mining extraction started")
        self._description = _(
            "A moon mining extraction has been started "
            "for %(structure_name)s at %(moon)s in %(solar_system)s "
            "belonging to %(owner_link)s. "
            "Extraction was started by %(character)s.\n"
            "The chunk will be ready on location at %(ready_time)s, "
            "and will fracture automatically on %(auto_time)s.\n"
            "\n%(ore_text)s"
        ) % {
            "structure_name": Webhook.text_bold(self._structure_name),
            "moon": self._moon.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "character": started_by,
            "ready_time": target_datetime_formatted(ready_time),
            "auto_time": target_datetime_formatted(auto_time),
            "ore_text": self.ore_text,
        }
        self._color = Webhook.Color.INFO


class NotificationMoonminningExtractionFinished(NotificationMoonminingEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        auto_time = ldap_time_2_datetime(self._data["autoTime"])
        self._title = _("Extraction finished")
        self._description = _(
            "The extraction for %(structure_name)s at %(moon)s "
            "in %(solar_system)s belonging to %(owner_link)s "
            "is finished and the chunk is ready "
            "to be shot at.\n"
            "The chunk will automatically fracture on %(auto_time)s.\n"
            "\n%(ore_text)s"
        ) % {
            "structure_name": Webhook.text_bold(self._structure_name),
            "moon": self._moon.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "auto_time": target_datetime_formatted(auto_time),
            "ore_text": self.ore_text,
        }
        self._color = Webhook.Color.INFO


class NotificationMoonminningAutomaticFracture(NotificationMoonminingEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = _("Automatic Fracture")
        self._description = _(
            "The moon drill fitted to %(structure_name)s at %(moon)s"
            " in %(solar_system)s belonging to %(owner_link)s "
            "has automatically been fired "
            "and the moon products are ready to be harvested.\n"
            "\n%(ore_text)s"
        ) % {
            "structure_name": Webhook.text_bold(self._structure_name),
            "moon": self._moon.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "ore_text": self.ore_text,
        }
        self._color = Webhook.Color.SUCCESS


class NotificationMoonminningExtractionCanceled(NotificationMoonminingEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        if self._data["cancelledBy"]:
            cancelled_by = get_or_create_eve_entity(id=self._data["cancelledBy"])
        else:
            cancelled_by = _("(unknown)")
        self._title = _("Extraction cancelled")
        self._description = _(
            "An ongoing extraction for %(structure_name)s at %(moon)s "
            "in %(solar_system)s belonging to %(owner_link)s "
            "has been cancelled by %(character)s."
        ) % {
            "structure_name": Webhook.text_bold(self._structure_name),
            "moon": self._moon.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "character": cancelled_by,
        }
        self._color = Webhook.Color.WARNING


class NotificationMoonminningLaserFired(NotificationMoonminingEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        fired_by = get_or_create_eve_entity(id=self._data["firedBy"])
        self._title = _("Moon drill fired")
        self._description = _(
            "The moon drill fitted to %(structure_name)s at %(moon)s "
            "in %(solar_system)s belonging to %(owner_link)s "
            "has been fired by %(character)s "
            "and the moon products are ready to be harvested.\n"
            "\n%(ore_text)s"
        ) % {
            "structure_name": Webhook.text_bold(self._structure_name),
            "moon": self._moon.name,
            "solar_system": self._solar_system_link,
            "owner_link": self._owner_link,
            "character": fired_by,
            "ore_text": self.ore_text,
        }
        self._color = Webhook.Color.SUCCESS



"""JSON serializers for Structures."""

# pylint: disable=missing-class-docstring

import re
from abc import ABC, abstractmethod
from typing import Optional

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.http import HttpRequest
from django.urls import reverse
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.core import dotlan
from eveuniverse.models import EvePlanet

from allianceauth.eveonline.models import EveCharacter
from app_utils.datetime import DATETIME_FORMAT, timeuntil_str
from app_utils.views import (
    BootstrapStyle,
    bootstrap_label_html,
    format_html_lazy,
    link_html,
    no_wrap_html,
    yesno_str,
    yesnonone_str,
)

from structures.app_settings import STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE
from structures.constants import EveGroupId
from structures.helpers import floating_icon_with_text_html
from structures.models import EveSpaceType, PocoDetails, Structure


class _AbstractStructureListSerializer(ABC):
    """Converting a list of structure objects into a dict for JSON."""

    ICON_RENDER_SIZE = 64

    def __init__(self, queryset: models.QuerySet, request: HttpRequest):
        self.queryset = queryset
        self._request = request

    def to_list(self) -> list:
        """Serialize all objects into a list."""
        return [self.serialize_object(obj) for obj in self.queryset]

    @abstractmethod
    def serialize_object(self, structure: Structure) -> dict:
        """Serialize one objects into a dict."""
        return {"id": structure.id}

    def _add_owner(self, structure: Structure, row: dict):
        corporation = structure.owner.corporation
        if corporation.alliance:
            alliance_name = corporation.alliance.alliance_name
            alliance_ticker = corporation.alliance.alliance_ticker
        else:
            alliance_name = alliance_ticker = ""

        if not structure.owner.is_structure_sync_fresh:
            update_warning_html = format_html(
                ' <i class="fas fa-exclamation-circle text-warning" '
                'title="Data has not been updated for a while and may be outdated."></i>'
            )
        else:
            update_warning_html = ""

        alliance_text = format_html(
            "<em>{}</em> {}", alliance_ticker, update_warning_html
        )
        owner_link = link_html(
            dotlan.corporation_url(corporation.corporation_name),
            corporation.corporation_name,
        )
        owner_display_html = floating_icon_with_text_html(
            corporation.logo_url(size=self.ICON_RENDER_SIZE),
            [owner_link, alliance_text],
        )
        row["owner"] = {
            "display": owner_display_html,
            "value": corporation.corporation_name,
        }
        row["alliance_name"] = (
            f"{alliance_name} [{alliance_ticker}]" if alliance_ticker else alliance_name
        )
        row["corporation_name"] = corporation.corporation_name

    def _add_location(self, structure: Structure, row: dict):
        solar_system = structure.eve_solar_system

        # location
        row["region_name"] = solar_system.eve_constellation.eve_region.name
        row["solar_system_name"] = solar_system.name
        solar_system_url = dotlan.solar_system_url(solar_system.name)
        if structure.eve_moon:
            location_name = structure.eve_moon.name
        elif structure.eve_planet:
            location_name = structure.eve_planet.name
        else:
            location_name = row["solar_system_name"]

        location_html = format_html(
            '<a href="{}">{}</a><br><em>{}</em>',
            solar_system_url,
            no_wrap_html(location_name),
            no_wrap_html(row["region_name"]),
        )
        row["location"] = {"display": location_html, "value": location_name}

    def _add_type(self, structure: Structure, row: dict):
        structure_type = structure.eve_type

        # category
        my_group = structure_type.eve_group
        row["group_name"] = my_group.name
        try:
            my_category = my_group.eve_category
            row["category_name"] = my_category.name
            row["is_starbase"] = structure.is_starbase
        except AttributeError:
            row["category_name"] = ""
            row["is_starbase"] = None

        # type
        type_link_html = link_html(structure_type.profile_url, structure_type.name)
        type_html = floating_icon_with_text_html(
            structure_type.icon_url(size=self.ICON_RENDER_SIZE),
            [type_link_html, format_html("<em>{}</em>", row["group_name"])],
        )
        row["type"] = {"display": type_html, "value": structure_type.name}
        row["type_name"] = structure_type.name

        # poco
        row["is_poco"] = structure.is_poco

    def _add_name_and_tags(
        self, structure: Structure, row: dict, check_tags: bool = True
    ):
        structure_name_html = escape(structure.name)
        tags = []
        if check_tags and structure.tags.exists():
            tags += [x.html for x in structure.tags.all()]
            structure_name_html += format_html("<br>{}", mark_safe(" ".join(tags)))

        row["structure_name_and_tags"] = structure_name_html

    def _add_reinforcement_infos(self, structure: Structure, row: dict):
        row["is_reinforced"] = structure.is_reinforced
        row["is_reinforced_str"] = yesno_str(structure.is_reinforced)

    def _add_fuel_and_power(self, structure: Structure, row: dict):
        fuel_expires_display, fuel_expires_timestamp = self._calc_fuel_infos(structure)
        last_online_at_display = self._calc_online_infos(structure)

        if fuel_expires_display or last_online_at_display:
            display = format_html(
                "{}<br>{}", no_wrap_html(fuel_expires_display), last_online_at_display
            )
        else:
            display = fuel_expires_display

        row["fuel_and_power"] = {
            "display": display,
            "fuel_expires_at": fuel_expires_timestamp,
        }
        row["power_mode_str"] = structure.get_power_mode_display()

    def _calc_fuel_infos(self, structure: Structure):
        if structure.is_poco:
            fuel_expires_display = ""
            fuel_expires_timestamp = None

        elif structure.is_low_power:
            fuel_expires_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.WARNING
                )
            )
            fuel_expires_timestamp = None

        elif structure.is_abandoned:
            fuel_expires_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.DANGER
                )
            )
            fuel_expires_timestamp = None

        elif structure.is_maybe_abandoned:
            fuel_expires_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.WARNING
                )
            )
            fuel_expires_timestamp = None

        elif structure.fuel_expires_at:
            fuel_expires_timestamp = structure.fuel_expires_at.isoformat()
            if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
                fuel_expires_display = timeuntil_str(
                    structure.fuel_expires_at - now(), show_seconds=False
                )
                if not fuel_expires_display:
                    fuel_expires_display = "?"
                    fuel_expires_timestamp = None
            else:
                if structure.fuel_expires_at >= now():
                    fuel_expires_display = structure.fuel_expires_at.strftime(
                        DATETIME_FORMAT
                    )
                else:
                    fuel_expires_display = "?"
                    fuel_expires_timestamp = None
        else:
            fuel_expires_display = "-"
            fuel_expires_timestamp = None
        return fuel_expires_display, fuel_expires_timestamp

    def _calc_online_infos(self, structure: Structure):
        if structure.is_poco:
            return ""

        if structure.is_full_power:
            last_online_at_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.SUCCESS
                )
            )
        elif structure.is_maybe_abandoned:
            last_online_at_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.WARNING
                )
            )

        elif structure.is_abandoned:
            last_online_at_display = format_html_lazy(
                bootstrap_label_html(
                    structure.get_power_mode_display(), BootstrapStyle.DANGER
                )
            )

        elif structure.last_online_at:
            if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
                last_online_at_display = timeuntil_str(
                    now() - structure.last_online_at, show_seconds=False
                )
                if not last_online_at_display:
                    last_online_at_display = "?"
            else:
                last_online_at_display = structure.last_online_at.strftime(
                    DATETIME_FORMAT
                )
        else:
            last_online_at_display = ""

        return last_online_at_display

    def _add_state_and_core(
        self, structure: Structure, row: dict, request: HttpRequest
    ):
        state_str, state_details = self._calc_state_infos(structure, request)
        core_status, has_core = self._calc_core_infos(structure)

        row["state_str"] = state_str
        row["state_details"] = format_html("{}<br>{}", state_details, core_status)
        row["core_status_str"] = yesnonone_str(has_core)

    def _calc_state_infos(self, structure: Structure, request):
        if structure.is_poco:
            return "-", "-"

        state_str = structure.get_state_display().capitalize()  # type: ignore
        state_details = format_html(state_str)
        if structure.state_timer_end:
            state_details += format_html(
                "<br>{}",
                no_wrap_html(structure.state_timer_end.strftime(DATETIME_FORMAT)),
            )

        if (
            request.user.has_perm("structures.view_all_unanchoring_status")
            and structure.unanchors_at
        ):
            state_details += format_html(
                "<br>Unanchoring until {}",
                no_wrap_html(structure.unanchors_at.strftime(DATETIME_FORMAT)),
            )

        return state_str, state_details

    def _calc_core_infos(self, structure: Structure):
        if structure.eve_type.eve_group_id not in {  # type: ignore
            EveGroupId.CITADEL,
            EveGroupId.ENGINEERING_COMPLEX,
            EveGroupId.REFINERY,
        }:
            return "", None

        if structure.has_core is True:
            has_core = True
            core_status = ""

        elif structure.has_core is False:
            has_core = False
            core_status = bootstrap_label_html("Core missing", BootstrapStyle.DANGER)

        else:
            has_core = None
            core_status = bootstrap_label_html("No core status", BootstrapStyle.WARNING)

        return core_status, has_core

    def _add_jump_fuel_level(self, structure: Structure, row: dict):
        if hasattr(structure, "jump_fuel_quantity_2"):
            row["jump_fuel_quantity"] = structure.jump_fuel_quantity_2  # type: ignore
        else:
            row["jump_fuel_quantity"] = None

    def _add_details_widget(
        self, structure: Structure, row: dict, request: HttpRequest
    ):
        """Add details widget when applicable"""
        if structure.has_fitting and request.user.has_perm(  # type: ignore
            "structures.view_structure_fit"
        ):
            ajax_url = reverse("structures:structure_details", args=[structure.id])
            row["details"] = format_html(
                '<button type="button" class="btn btn-default" '
                'data-toggle="modal" data-target="#modalUpwellDetails" '
                f"data-ajax_url={ajax_url} "
                f'title="{_("Show fitting")}">'
                '<i class="fas fa-search"></i></button>'
            )

        elif structure.is_poco:
            ajax_url = reverse("structures:poco_details", args=[structure.id])
            row["details"] = format_html(
                '<button type="button" class="btn btn-default" '
                'data-toggle="modal" data-target="#modalPocoDetails" '
                f"data-ajax_url={ajax_url} "
                f'title="{_("Show details")}">'
                '<i class="fas fa-search"></i></button>'
            )

        elif structure.is_starbase:
            ajax_url = reverse("structures:starbase_detail", args=[structure.id])
            row["details"] = format_html(
                '<button type="button" class="btn btn-default" '
                'data-toggle="modal" data-target="#modalStarbaseDetail" '
                f"data-ajax_url={ajax_url} "
                f'title="{_("Show details")}">'
                '<i class="fas fa-search"></i></button>'
            )
        else:
            row["details"] = ""

    @staticmethod
    def extract_planet_type_name(eve_planet: EvePlanet) -> str:
        """Extract short name of planet type."""
        matches = re.findall(r"Planet \((\S*)\)", eve_planet.eve_type.name)
        return matches[0] if matches else ""


class StructureListSerializer(_AbstractStructureListSerializer):
    def __init__(self, queryset: models.QuerySet, request: HttpRequest):
        super().__init__(queryset, request=request)
        self.queryset = self.queryset.prefetch_related("tags")

    def serialize_object(self, structure: Structure) -> dict:
        row = super().serialize_object(structure)
        self._add_owner(structure, row)
        self._add_location(structure, row)
        self._add_type(structure, row)
        self._add_name_and_tags(structure, row)
        self._add_reinforcement_infos(structure, row)
        self._add_fuel_and_power(structure, row)
        self._add_state_and_core(structure, row, self._request)
        self._add_details_widget(structure, row, self._request)
        self._add_jump_fuel_level(structure, row)
        return row


class PocoListSerializer(_AbstractStructureListSerializer):
    def __init__(
        self,
        queryset: models.QuerySet,
        request: HttpRequest,
        character: Optional[EveCharacter] = None,
    ):
        super().__init__(queryset, request=request)
        self.queryset = self.queryset.select_related(
            "eve_planet",
            "eve_planet__eve_type",
            "eve_planet__eve_type__eve_group",
            "eve_type",
            "eve_type__eve_group",
            "eve_solar_system",
            "eve_solar_system__eve_constellation__eve_region",
            "poco_details",
            "owner__corporation",
            "owner__corporation__alliance",
        )
        if not request:
            raise ValueError("request can not be None")

        self.character = character

    def serialize_object(self, structure: Structure) -> dict:
        row = super().serialize_object(structure)
        self._add_owner(structure, row)
        self._add_location(structure, row)
        self._add_planet(structure, row)
        self._add_has_access_and_tax(structure, row, self.character)
        return row

    def _add_location(self, structure: Structure, row: dict):
        if structure.eve_solar_system.is_low_sec:
            space_badge_type = "warning"
        elif structure.eve_solar_system.is_high_sec:
            space_badge_type = "success"
        else:
            space_badge_type = "danger"
        space_type = EveSpaceType.from_solar_system(structure.eve_solar_system)

        solar_system_name = structure.eve_solar_system.name
        solar_system_html = format_html(
            "{}<br>{}",
            link_html(dotlan.solar_system_url(solar_system_name), solar_system_name),
            bootstrap_label_html(text=space_type.label, label=space_badge_type),
        )

        constellation_name = structure.eve_solar_system.eve_constellation.name
        region_name = structure.eve_solar_system.eve_constellation.eve_region.name
        constellation_html = format_html(
            "{}<br><em>{}</em>", constellation_name, region_name
        )

        row["constellation_html"] = {
            "display": constellation_html,
            "sort": constellation_name,
        }
        row["solar_system_html"] = {
            "display": solar_system_html,
            "sort": solar_system_name,
        }
        row["solar_system"] = solar_system_name
        row["constellation"] = f"{region_name} - {constellation_name}"
        row["region"] = region_name
        row["space_type"] = space_type.value

    def _add_planet(self, structure: Structure, row: dict):
        if structure.eve_planet:
            planet_type_name = self.extract_planet_type_name(structure.eve_planet)
            planet_name = structure.eve_planet.name
            icon_url = structure.eve_planet.eve_type.icon_url(
                size=self.ICON_RENDER_SIZE
            )
        else:
            planet_name = planet_type_name = "?"
            icon_url = ""

        planet_plus_icon_html = floating_icon_with_text_html(
            icon_url, [planet_name, format_html("<em>{}</em>", planet_type_name)]
        )

        row["planet_plus_icon"] = {
            "display": planet_plus_icon_html,
            "sort": planet_name,
        }
        row["planet_type_name"] = planet_type_name
        row["planet_name"] = planet_name

    def _add_has_access_and_tax(
        self, structure: Structure, row: dict, character: Optional[EveCharacter]
    ):
        if character:
            access_info = self._determine_access_and_tax_for_character(
                structure, character
            )
        else:
            access_info = None

        has_access_html = ""
        has_access_str = "?"
        tax = None
        tax_html = "?"
        if not access_info:
            has_access_html = format_html_lazy(
                '<i class="fas fa-question" title="{}"></i>',
                _("Unknown if character has access"),
            )

        else:
            dubious_access_text = _(
                "Access and tax for characters, which are not a member of "
                "the owner corporation or it's alliance may not be accurate."
            )

            if not access_info.has_access:
                has_access_html = format_html_lazy(
                    '<i class="fas fa-times text-danger text-tooltip" title="{}"></i>',
                    _("No access"),
                )
                has_access_str = _("no")

            elif access_info.has_access:
                if access_info.is_confident:
                    has_access_html = format_html_lazy(
                        '<i class="fas fa-check text-success text-tooltip" title="{}"></i>',
                        _("Has access"),
                    )
                    has_access_str = _("yes")

                else:
                    has_access_html = format_html_lazy(
                        '<span class="text-tooltip" title="{}">'
                        '(<i class="fas fa-check text-success text-tooltip"></i>)</span>',
                        dubious_access_text,
                    )
                    has_access_str = _("yes(?)")

            if access_info.has_access and access_info.tax_rate is not None:
                tax = access_info.tax_rate * 100
                tax_str = f"{tax:.0f} %"
                if not access_info.is_confident:
                    tax_html = format_html_lazy(
                        '<span class="text-tooltip" title="{}">{}</span>',
                        dubious_access_text,
                        tax_str,
                    )
                else:
                    tax_html = tax_str
            else:
                tax = None
                tax_html = "?"

        row["has_access_html"] = {"display": has_access_html, "sort": has_access_str}
        row["has_access_str"] = has_access_str
        row["tax"] = {"display": tax_html, "sort": tax}

    def _determine_access_and_tax_for_character(
        self, structure, character
    ) -> Optional[PocoDetails.PocoCharacterAccessInfo]:
        try:
            details: PocoDetails = structure.poco_details  # type: ignore
        except (AttributeError, ObjectDoesNotExist):
            return None

        return details.determine_access_and_tax_for_character(character)



from unittest.mock import Mock, patch

from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from structures.core import notification_timers
from structures.core.notification_types import NotificationType
from structures.models import Notification
from structures.tests.testdata.factories import (
    GeneratedNotificationFactory,
    OwnerFactory,
    RefineryFactory,
    StructureFactory,
)
from structures.tests.testdata.helpers import (
    load_eve_entities,
    load_notification_entities,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer

    MODULE_PATH = "structures.core.notification_timers"

    @patch(
        "structuretimers.models._task_calc_timer_distances_for_all_staging_systems",
        Mock(),
    )
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    class TestNotificationAddToTimerboard(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            load_eveuniverse()
            load_eve_entities()
            cls.owner = OwnerFactory(is_alliance_main=True)
            load_notification_entities(cls.owner)
            StructureFactory(owner=cls.owner, id=1000000000001)
            RefineryFactory(owner=cls.owner, id=1000000000002)

        @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", False)
        @patch("allianceauth.timerboard.models.Timer", spec=True)
        def test_moon_timers_disabled(self, mock_Timer):
            # given
            notif = Notification.objects.get(notification_id=1000000404)
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertFalse(result)
            self.assertFalse(mock_Timer.objects.create.called)
            notif.refresh_from_db()
            self.assertFalse(notif.add_or_remove_timer())
            self.assertFalse(mock_Timer.delete.called)

        @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
        def test_normal(self):
            notification_without_timer_query = Notification.objects.filter(
                notification_id__in=[
                    1000000401,
                    1000000403,
                    1000000405,
                    1000000502,
                    1000000503,
                    1000000506,
                    1000000507,
                    1000000508,
                    1000000509,
                    1000000510,
                    1000000511,
                    1000000512,
                    1000000513,
                    1000000601,
                    1000010509,
                    1000010601,
                ]
            )
            for obj in notification_without_timer_query:
                self.assertFalse(obj.add_or_remove_timer())

            self.assertEqual(AuthTimer.objects.count(), 0)

            obj = Notification.objects.get(notification_id=1000000501)
            self.assertFalse(obj.add_or_remove_timer())

            obj = Notification.objects.get(notification_id=1000000504)
            self.assertTrue(obj.add_or_remove_timer())

            obj = Notification.objects.get(notification_id=1000000505)
            self.assertTrue(obj.add_or_remove_timer())

            obj = Notification.objects.get(notification_id=1000000602)
            self.assertTrue(obj.add_or_remove_timer())

            ids_set_1 = {x.id for x in AuthTimer.objects.all()}
            obj = Notification.objects.get(notification_id=1000000404)
            self.assertTrue(obj.add_or_remove_timer())

            self.assertEqual(AuthTimer.objects.count(), 4)

            # this should remove the right timer only
            obj = Notification.objects.get(notification_id=1000000402)
            obj.add_or_remove_timer()
            self.assertEqual(AuthTimer.objects.count(), 3)
            ids_set_2 = {x.id for x in AuthTimer.objects.all()}
            self.assertSetEqual(ids_set_1, ids_set_2)

        @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
        def test_run_all(self):
            for obj in Notification.objects.all():
                timer_types = NotificationType.relevant_for_timerboard()
                with self.subTest(notif_type=obj.notif_type):
                    is_timer = obj.notif_type in timer_types
                    is_added = obj.add_or_remove_timer()
                    self.assertEqual(is_timer, is_added)

        @patch(MODULE_PATH + ".STRUCTURES_TIMERS_ARE_CORP_RESTRICTED", False)
        def test_corp_restriction_1(self):
            # given
            notif = Notification.objects.get(notification_id=1000000504)
            # when
            result = notif.add_or_remove_timer()
            # then
            self.assertTrue(result)
            timer = AuthTimer.objects.first()
            self.assertFalse(timer.corp_timer)

        @patch(MODULE_PATH + ".STRUCTURES_TIMERS_ARE_CORP_RESTRICTED", True)
        def test_corp_restriction_2(self):
            obj = Notification.objects.get(notification_id=1000000504)
            self.assertTrue(obj.add_or_remove_timer())
            timer = AuthTimer.objects.first()
            self.assertTrue(timer.corp_timer)

        def test_timer_starbase_reinforcement(self):
            # given
            notif = GeneratedNotificationFactory()
            structure = notif.structures.first()
            # when
            result = notif.add_or_remove_timer()
            # then
            self.assertTrue(result)
            obj = AuthTimer.objects.first()
            self.assertEqual(obj.system, structure.eve_solar_system.name)
            self.assertEqual(obj.planet_moon, structure.eve_moon.name)
            self.assertEqual(obj.eve_time, structure.state_timer_end)
